'''This module contains a class for each use case'''

import math
import os
from subprocess import *

from Ganga.Core import ApplicationConfigurationError
from Ganga.Core import GangaException
from Ganga.GPIDev.Schema import *
from Ganga.GPIDev.Lib.Dataset import Dataset
from Ganga.GPIDev.Lib.File import File
import Ganga.Utility.logging

import db
import SBDatasetManager
import utils


logger = Ganga.Utility.logging.getLogger()


class SBInputProductionAnalysis(Dataset):
    '''Input dataset class for (official or personal) production data analysis'''
    
    _name = "SBInputProductionAnalysis"
    _category = "datasets"
    _exportmethods = ["check",
                      "getDataset"]
    _schema = Schema( Version( 1, 0 ), {
        # not protected
        "dataset_id" : SimpleItem(defvalue='', hidden=0, protected=0, copyable=1, doc="Input dataset_id"),
        "events_total" : SimpleItem(defvalue=0, hidden=0, protected=0, copyable=1, doc="minumun number of events to analyze, zero for all"),
        "events_per_subjobs" : SimpleItem(defvalue=0, hidden=0, protected=0, copyable=1, doc="maximum number of events for each subjob"),
        # protected
        "input_path" : SimpleItem(defvalue=[], hidden=0, protected=1, copyable=1, doc="List of input files"),
        "input_mode" : SimpleItem(defvalue='list', hidden=0, protected=1, copyable=1, doc="Input mode"),
        "number_of_subjobs" : SimpleItem(defvalue=0, protected=1, copyable=1, doc="Number of Subjobs"),
        "run_site" : SimpleItem(defvalue=[], protected=1, copyable=1, doc="Site where job will be running"),
    })
    
    def check(self):
        '''Checks done during submit phase. Protected elements are written.'''
        
        if self.dataset_id == '':
            raise ApplicationConfigurationError(None, 'You must define an input dataset')
        
        if self.events_per_subjobs == 0:
            raise ApplicationConfigurationError(None, 'You must define events_per_subjobs')
        
        kwargs = dict()
        kwargs['dataset_id'] = self.dataset_id
        
        manager = SBDatasetManager.SBDatasetManager()
        datasets = manager.getDataset(**kwargs)
        
        # only one dataset
        if len(datasets) == 0:
            msg = 'Input dataset %s not found' % self.dataset_id
            raise ApplicationConfigurationError(None, msg)
        assert len(datasets) == 1, 'Dataset consistency error'
        dataset = datasets[0]
        
        # status
        if dataset['status'] not in ['open', 'closed']:
            msg = 'Input dataset %s status is not open or closed' % self.dataset_id
            raise ApplicationConfigurationError(None, msg)
        
        # session
        if dataset['session'] not in ['fastsim', 'fullsim']:
            msg = 'Input dataset %s session is not fastsim or fullsim' % self.dataset_id
            raise ApplicationConfigurationError(None, msg)
        
        if self.events_total == 0:
            self.events_total = int(dataset['parameters']['evt_tot'])
        
        if self.events_total != 0 and self.events_total > int(dataset['parameters']['evt_tot']):
            msg = 'Input dataset %s total events is %d' % (self.dataset_id, dataset['parameters']['evt_tot'])
            raise ApplicationConfigurationError(None, msg)
        
        if self.events_per_subjobs < int(dataset['parameters']['evt_file']):
            msg = 'events_per_subjobs must be >= %s' % dataset['parameters']['evt_file']
            raise ApplicationConfigurationError(None, msg)
        
        if self.events_per_subjobs >= self.events_total:
            msg = 'events_per_subjobs cannot be >= events_total'
            raise ApplicationConfigurationError(None, msg)
        
        lfns = self.__getLFNs(dataset['parameters']['evt_file'])
        self.__createInputPath(lfns, dataset['parameters']['evt_file'])
        self.input_mode = 'list'
    
    def __getLFNs(self, evt_file):
        '''This method returns a list of LFNs from sbk database for the 
        selected dataset_id. The number of LFNs is calculated from the 
        number of selected events (events_total).'''
        
        if self.events_total == 0:
            num_files = 0
        else:
            num_files = int(math.ceil(float(self.events_total)/int(evt_file)))
        
        sql = 'SELECT lfn, size FROM output_union WHERE dataset_id = %s'
        param = [r'\x' + self.dataset_id]
        
        if num_files > 0:
            sql += ' ORDER BY RANDOM() LIMIT %s'
            param.append(num_files)
        
        return db.read(sql, param)
    
    def __createInputPath(self, lfns, evt_file):
        '''This method splits the list of LFNs between subjobs and writes a 
        text file for each one.'''
        
        evt_file = int(evt_file)
        
        # split all lfns between subjobs
        job = list()
        job.append(list())
        size = 0
        events = 0
        maxInput = 10 * (2**30) # 10GiB
        minInput = 2 * (2**30) # 2GiB
        
        # fill the subjobs al long as there are LFNs,
        # to determine the number of subjobs required 
        for lfn in lfns:
            if (size + int(lfn['size'])) < maxInput and (events + evt_file) <= self.events_per_subjobs:
                size += int(lfn['size'])
                events += evt_file
            else:
                job.append(list())
                size = int(lfn['size'])
                events = evt_file
            
            job[-1].append(lfn)
        
        self.number_of_subjobs = len(job)
        
        # level the number of LFNs between the subjob.
        tot_files = len(lfns)
        balanced_number_lfn_per_subjob = int(math.ceil(float(tot_files)/self.number_of_subjobs))
        job = list()
        self.input_path = list()
        max_size = 0
        jobInputDir = self.getJobObject().inputdir
        lfns_index = 0
        
        for subjob_id in xrange(self.number_of_subjobs):
            subjob = dict()
            size = 0
            events = 0
            number_lfns = 0
            subjob['id'] = str(subjob_id)
            subjob['list_path'] = os.path.join(jobInputDir, "list_%d.txt" % subjob_id)
            
            f = open(subjob['list_path'], 'w')
            try:
                for lfn in lfns[lfns_index:lfns_index + balanced_number_lfn_per_subjob]:
                    f.write(lfn['lfn'] + '\n')
                    size += int(lfn['size'])
                    events += evt_file
                    number_lfns += 1
            finally:
                f.close()
            
            lfns_index += balanced_number_lfn_per_subjob
            self.input_path.append(File(f.name))
            subjob['size'] = utils.sizeof_fmt_binary(size)
            subjob['events'] = utils.sizeof_fmt_decimal(int(events))
            subjob['lfns'] = number_lfns
            job.append(subjob)
            
            if size > max_size:
                max_size = size
        
        if max_size < minInput:
            logger.warning('These subjobs input is very small, to improve the \
            efficiency you could increase the numbers of events per subjob.')
        
        return job
    
    def getDefaultRunSite(self):
        '''This method is called only if backend is LCG.
        It retrieves the sites associated to dataset_id.'''
        
        sql = 'SELECT site FROM dataset_site_union WHERE dataset_id = %s'
        sites = db.read(sql, (r'\x' + self.dataset_id, ))
        
        for site in sites:
            self.run_site.append(site['site'])
    
    def getDataset(self, **kwargs):
        '''Interactive mathod. It prints the datasets (the user can apply filters),
        the user chooses one of them and inserts the number of events he wants.'''
        
        manager = SBDatasetManager.SBDatasetManager()
        
        def validateFilter(filter, allowed):
            kwargs[filter] = kwargs.get(filter, allowed)
            if not isinstance(kwargs[filter], list):
                kwargs[filter] = [kwargs[filter]]
            if not set(kwargs[filter]).issubset(set(allowed)):
                raise GangaException('%s must be %s' % (filter, allowed))
        
        validateFilter('status', ['open', 'closed'])
        validateFilter('session', ['fastsim', 'fullsim'])
        
        datasets = manager.getDataset(**kwargs)
        dataset = manager.printDatasets(datasets)
        
        self.dataset_id = dataset['dataset_id']
        
        print('\nChosen dataset details:')
        manager.printDatasetDetail(dataset)
        
        print('\nInsert the minimum number of events that you need for your analysis (zero for all):')
        self.events_total = utils.getIndex(maxInclusive=int(dataset['parameters']['evt_tot']))
        
        lfns = self.__getLFNs(dataset['parameters']['evt_file'])
        
        tot_size = 0
        tot_files = len(lfns)
        tot_events = int(dataset['parameters']['evt_file']) * tot_files
        
        for lfn in lfns:
            tot_size += int(lfn['size'])
        
        print('\nTotal job input size: ' + str(utils.sizeof_fmt_binary(tot_size)))
        print('Total selected number of events: ' + str(utils.sizeof_fmt_decimal(tot_events)))
        print('Total number of involved lfns: ' + str(tot_files))
        
        print('\nInsert the maximum number of events for each subjob. Remember:')
        print('- maximum output size is 2GiB.')
        print('- suggested maximum job duration 18h.')
        print('- maximum input size job is 10GiB.')
        print('- at least %s (that is the number of events of one file).' % dataset['parameters']['evt_file'])
        
        self.events_per_subjobs = utils.getIndex(minInclusive=int(dataset['parameters']['evt_file']), maxInclusive=tot_events)
        job = self.__createInputPath(lfns, dataset['parameters']['evt_file'])
        
        print('\nSubjobs details:')
        column_names = ['id', 'list_path', 'size', 'events', 'lfns']
        print(utils.format_dict_table(job, column_names))




class SBInputPureAnalysis(Dataset):
    '''Input dataset class that manage input dataset coming from use case analysis'''
    
    _name = "SBInputPureAnalysis"
    _category = "datasets"
    _exportmethods = ["check",
                      "getDataset"]
    _schema = Schema( Version( 1, 0 ), {
        # not protected
        "dataset_id" : SimpleItem(defvalue='', hidden=0, protected=0, copyable=1, doc="Input dataset_id"),
        "files_total" : SimpleItem(defvalue=0, hidden=0, protected=0, copyable=1, doc="minumun number of file to analyze, zero for all"),
        "files_per_subjobs" : SimpleItem(defvalue=0, hidden=0, protected=0, copyable=1, doc="maximum number of file for each subjob"),
        # protected
        "input_path" : SimpleItem(defvalue=[], hidden=0, protected=1, copyable=1, doc="List of input files"),
        "input_mode" : SimpleItem(defvalue='list', hidden=0, protected=1, copyable=1, doc="Input mode"),
        "number_of_subjobs" : SimpleItem(defvalue=0, protected=1, copyable=1, doc="Number of Subjobs"),
        "run_site" : SimpleItem(defvalue=[], protected=1, copyable=1, doc="Site where job will be running"),
    })
    
    def check(self):
        '''Checks done during submit phase. Protected elements are written.'''
        
        if self.dataset_id == '':
            raise ApplicationConfigurationError(None, 'You must define an input dataset')
        
        if self.files_per_subjobs == 0:
            raise ApplicationConfigurationError(None, 'You must define events_per_subjobs')
        
        kwargs = dict()
        kwargs['dataset_id'] = self.dataset_id
        
        manager = SBDatasetManager.SBDatasetManager()
        datasets = manager.getDataset(**kwargs)
        
        # only one dataset
        if len(datasets) == 0:
            msg = 'Input dataset %s not found' % self.dataset_id
            raise ApplicationConfigurationError(None, msg)
        assert len(datasets) == 1, 'Dataset consistency error'
        dataset = datasets[0]
        
        # status
        if dataset['status'] not in ['open', 'closed']:
            msg = 'Input dataset %s status is not open or closed' % self.dataset_id
            raise ApplicationConfigurationError(None, msg)
        
        # session
        if dataset['session'] not in ['analysis']:
            msg = 'Input dataset %s is not analysis' % self.dataset_id
            raise ApplicationConfigurationError(None, msg)
        
        if self.files_total == 0:
            self.files_total = int(dataset['files'])
        
        if self.files_total != 0 and self.files_total > int(dataset['files']):
            msg = 'Input dataset %s total files is %d' % (self.dataset_id, dataset['files'])
            raise ApplicationConfigurationError(None, msg)
        
        if self.files_per_subjobs >= self.files_total:
            msg = 'files_per_subjobs cannot be >= files_total'
            raise ApplicationConfigurationError(None, msg)
        
        lfns = self.__getLFNs()
        self.__createInputPath(lfns)
        self.input_mode = 'list'
    
    def __getLFNs(self):
        '''This method returns a list of LFNs from sbk database for the 
        selected dataset_id. The number of LFNs is provided by user.'''
        
        sql = 'SELECT lfn, size FROM output_union WHERE dataset_id = %s'
        param = [r'\x' + self.dataset_id]
        
        if self.files_total > 0:
            sql += ' ORDER BY RANDOM() LIMIT %s'
            param.append(self.files_total)
        
        return db.read(sql, param)
    
    def __createInputPath(self, lfns):
        '''This method splits the list of LFNs between subjobs and writes a 
        text file for each one.'''
        
        # split all lfns between subjobs
        job = list()
        job.append(list())
        size = 0
        files = 0
        maxInput = 10 * (2**30) # 10GiB
        minInput = 2 * (2**30) # 2GiB
        
        # fill the subjobs al long as there are LFNs,
        # to determine the number of subjobs required 
        for lfn in lfns:
            if (size + int(lfn['size'])) < maxInput and (files + 1) <= self.files_per_subjobs:
                size += int(lfn['size'])
                files += 1
            else:
                job.append(list())
                size = int(lfn['size'])
                files = 1
            
            job[-1].append(lfn)
        
        self.number_of_subjobs = len(job)
        
        # level the number of LFNs between the subjob.
        tot_files = len(lfns)
        balanced_number_lfn_per_subjob = int(math.ceil(float(tot_files)/self.number_of_subjobs))
        job = list()
        self.input_path = list()
        max_size = 0
        jobInputDir = self.getJobObject().inputdir
        lfns_index = 0
        
        for subjob_id in xrange(self.number_of_subjobs):
            subjob = dict()
            size = 0
            events = 0
            number_lfns = 0
            subjob['id'] = str(subjob_id)
            subjob['list_path'] = os.path.join(jobInputDir, "list_%d.txt" % subjob_id)
            
            f = open(subjob['list_path'], 'w')
            try:
                for lfn in lfns[lfns_index:lfns_index + balanced_number_lfn_per_subjob]:
                    f.write(lfn['lfn'] + '\n')
                    size += int(lfn['size'])
                    number_lfns += 1
            finally:
                f.close()
            
            lfns_index += balanced_number_lfn_per_subjob
            self.input_path.append(File(f.name))
            subjob['size'] = utils.sizeof_fmt_binary(size)
            subjob['lfns'] = number_lfns
            job.append(subjob)
            
            if size > max_size:
                max_size = size
        
        if max_size < minInput:
            logger.warning('These subjobs input is very small, to improve the efficiency you could increase the numbers of events per subjob.')
        
        return job
    
    def getDefaultRunSite(self):
        '''This method is called only if backend is LCG.
        It retrieves the sites associated to dataset_id.'''
        
        sql = 'SELECT site FROM dataset_site_union WHERE dataset_id = %s'
        sites = db.read(sql, (r'\x' + self.dataset_id, ))
        
        for site in sites:
            self.run_site.append(site['site'])
    
    def getDataset(self, **kwargs):
        '''Interactive method. It prints the datasets (the user can apply filters),
        the user chooses one of them and inserts the number of LFNs he wants.'''
        
        manager = SBDatasetManager.SBDatasetManager()
        
        def validateFilter(filter, allowed):
            kwargs[filter] = kwargs.get(filter, allowed)
            if not isinstance(kwargs[filter], list):
                kwargs[filter] = [kwargs[filter]]
            if not set(kwargs[filter]).issubset(set(allowed)):
                raise GangaException('%s must be %s' % (filter, allowed))
        
        validateFilter('status', ['open', 'closed'])
        validateFilter('session', ['analysis'])
        kwargs['files'] = 0
        
        datasets = manager.getDataset(**kwargs)
        dataset = manager.printDatasets(datasets)
        
        self.dataset_id = dataset['dataset_id']
        
        print('\nChosen dataset details:')
        manager.printDatasetDetail(dataset)
        
        print('\nInsert the minimum number of files that you need for your analysis (zero for all):')
        self.files_total = utils.getIndex(maxInclusive=int(dataset['files']))
        
        lfns = self.__getLFNs()
        
        tot_size = 0
        tot_files = len(lfns)
        
        for lfn in lfns:
            tot_size += int(lfn['size'])
        
        print('\nTotal job input size: ' + str(utils.sizeof_fmt_binary(tot_size)))
        print('Total number of involved lfns: ' + str(tot_files))
        
        print('\nInsert the maximum number of files for each subjob. Remember:')
        print('- maximum output size is 2GiB.')
        print('- suggested maximum job duration 18h.')
        print('- maximum input size job is 10GiB.')
        
        self.files_per_subjobs = utils.getIndex(minInclusive=1, maxInclusive=tot_files)
        job = self.__createInputPath(lfns)
        
        print('\nSubjobs details:')
        column_names = ['id', 'list_path', 'size', 'lfns']
        print(utils.format_dict_table(job, column_names))




class SBInputPersonalProduction(Dataset):
    '''Input dataset class for personal production use case. If session is:
    - fullsim: no input is requiered.
    - fastsim: user can enable last approved background frame as job input.'''
    
    _name = "SBInputPersonalProduction"
    _category = "datasets"
    _exportmethods = ["check",
                      "interactive",
                      "setSoftwareVersion"
                      ]
    _schema = Schema( Version( 1, 0 ), {
        "input_path" : SimpleItem(defvalue=[], hidden=0, protected=1, copyable=1, doc="List of input files"),
        "input_mode" : SimpleItem(defvalue='none', hidden=0, protected=1, copyable=1, doc="Input mode"),
        "run_site" : SimpleItem(defvalue=[], protected=1, copyable=1, doc="Site where job will be running"),
        
        "background_frame" : SimpleItem(defvalue=False, protected=0, copyable=1, doc="Enable input background frame if session is FastSim"),
        "number_of_subjobs" : SimpleItem(defvalue=0, doc="Number of Subjobs (1-250"),
        "soft_version" : SimpleItem(defvalue="", protected=0, copyable=1, doc="Simulation software version. Can be empty."),
        "session" : SimpleItem(defvalue="", protected=0, copyable=1, doc="Simulation type name. Can be empty."),
    })
    
    def __sbcurrent(self):
        '''Reads file .sbcurrent in user software directory to find session
        name and software version. If it can not find them, throws an exception.'''
        
        j = self.getJobObject()
        
        line = None
        
        try:
            sbcurrent = os.path.join(j.application.software_dir, '.sbcurrent')
            f = open(sbcurrent)
            try:
                line = f.read().splitlines()[0]
            finally:
                f.close()
        except:
            pass
        
        if line is None or len(line) == 0:
            raise GangaException('Unable to find software and the version in .sbcurrent')
        
        (self.session, self.soft_version) = line.split('/')
        logger.info('Found in .sbcurrent: %s %s' % (self.session, self.soft_version))
    
    def check(self):
        '''Checks done during submit phase. Protected elements are written.'''
        
        # controllare che session sia FastSim o FullSim
        if self.soft_version == '' and self.session == '':
            self.__sbcurrent()
        
        allowed_session = ['FastSim', 'FullSim']
        if self.session not in allowed_session:
            raise ApplicationConfigurationError(None, 'session must be %s' % allowed_session)
        
        if self.number_of_subjobs < 1 or self.number_of_subjobs > 250:
            raise ApplicationConfigurationError(None, 'number_of_subjobs must be between 1 and 250')
        
        sql = '''SELECT DISTINCT soft_version
            FROM session_site_soft
            WHERE session_name = %s
            ORDER BY soft_version'''
        
        supported_soft_version = db.read(sql, (self.session, ))
        # convert from list of dict to list of strings
        supported_soft_version = [s['soft_version'] for s in supported_soft_version]
        
        if self.soft_version not in supported_soft_version:
            raise ApplicationConfigurationError(None, 'supported soft_version are: %s' % supported_soft_version)
        
        if self.session == 'FastSim' and self.background_frame == True:
            results = db.read('''SELECT prod_series, lfn_dir
                FROM background_frame
                WHERE valid = true
                ORDER BY validation_timestamp DESC
                LIMIT 1''')
            
            self.input_path = list()
            self.input_path.append(results[0]['lfn_dir'])
            self.input_mode = 'dir'
            self.background_frame_prod_series = results[0]['prod_series']
            logger.info('Last approved (%s) background frame has been setup \
            for job input' % self.background_frame_prod_series)
    
    def interactive(self):
        '''Interactive find software version inside user software directory'''
        
        try:
            self.__sbcurrent()
        except GangaException as e:
            logger.debug(e)
            self.setSoftwareVersion()
    
    def setSoftwareVersion(self):
        '''Set software version. This function is called when interactive() 
        cannot find it automatically or is called by user if prefer set it 
        manually'''
        
        results = db.read('''SELECT session_name, soft_version
            FROM session_site_soft
            GROUP BY session_name, soft_version
            ORDER BY session_name, soft_version''')
        
        i = 0
        for result in results:
            result['id'] = i
            i += 1
        
        print('Choose simulation type:')
        column_names = ('id', 'session_name', 'soft_version')
        print(utils.format_dict_table(results, column_names))
        
        index = utils.getIndex(maxExclusive=len(results))
        
        self.session = results[index]['session_name']
        self.soft_version = results[index]['soft_version']
    
    def getDefaultRunSite(self):
        '''This method is called only if backend is LCG.
        It populates the run_site list with all compatible sites.'''
        
        sql = '''SELECT site
            FROM session_site_soft
            WHERE session_name = %s AND soft_version = %s'''
        param = [self.session, self.soft_version]
        
        if self.background_frame:
            sql += ''' AND site IN (
                SELECT site
                FROM background_frame_site
                WHERE prod_series = %s)'''
            param.append(self.background_frame_prod_series)
        
        results = db.read(sql, param)
        
        if len(results) == 0:
            raise ApplicationConfigurationError(None, 'No site found with the specified requirements')
        
        for result in results:
            self.run_site.append(result['site'])




from Ganga.Core.exceptions import SplitterError
from Ganga.GPIDev.Adapters.ISplitter import SplittingError
from GangaDirac.Lib.Backends.DiracUtils import result_ok
from Ganga.Utility.Config import getConfig
from Ganga.Utility.logging import getLogger
from copy import deepcopy
import random

configDirac = getConfig('DIRAC')
logger = getLogger()

global_random = random

LFN_parallel_limit = 250
limit_divide_one = 1. / float(LFN_parallel_limit)

# Find a random element from a python list that isn't in a given banned list
# This is used for selecting a site element but it doesn't matter which is used


def find_random_site(original_SE_list, banned_SE):
    input_list = deepcopy(original_SE_list)
    chosen_element = ""

    while chosen_element == "" and len(input_list) > 0:
        global global_random
        this_element = global_random.sample(input_list, 1)[0]
        if not this_element in banned_SE:
            chosen_element = this_element
            break
        else:
            input_list.remove(this_element)

    return chosen_element

# This function is used for adding all of the known site for a given SE
# The output is stored in the dictionary site_to_SE_mapping


def addToMapping(SE, site_to_SE_mapping):

    from GangaDirac.Lib.Utilities.DiracUtilities import execute
    result = execute('getSitesForSE("%s")' % str(SE))
    if result.get('OK') != True:
        logger.error("Error getting SE's for site: %s" % str(SE))
        site_to_SE_mapping[SE] = []
    else:
        usable_site = result.get('Value')
        site_to_SE_mapping[SE] = usable_site


def getLFNReplicas(allLFNs, index, allLFNData):

    output = None

    toy_num = 0

    global LFN_parallel_limit

    this_min = index * LFN_parallel_limit

    if (index + 1) * LFN_parallel_limit > len(allLFNs):
        this_max = len(allLFNs)
    else:
        this_max = (index + 1) * LFN_parallel_limit

    for toy_num in range(5):

        try:
            from GangaDirac.Lib.Utilities.DiracUtilities import execute
            output = execute('getReplicas(%s)' % str(allLFNs[this_min:this_max]))
            these_values = output.get('Value').get('Successful')
            break
        except Exception, err:
            logger.error("Dirac Error: %s" % str(err))
            # catch 'Successful' not found and others
            pass

    import Ganga.Runtime.Repository_runtime
    Ganga.Runtime.Repository_runtime.updateLocksNow()

    allLFNData[index] = output

    if toy_num == 4:
        logger.error("Failed to Get Replica Info: [%s:%s] of %s" (
            str(this_min), str(this_max), len(allLFNs)))
    else:
        logger.info("Got Replica Info: [%s:%s] of %s" % (
            str(this_min), str(this_max), len(allLFNs)))

# For a given site select 'wanted_common_site' many site from a given Set
# If uniqueSE is requestd the site selected will NOT share a common SE


def generate_site_selection(input_site, wanted_common_site, uniqueSE, site_to_SE_mapping, SE_to_site_mapping):
    req_sitez = set([])
    if len(input_site) > wanted_common_site:
        used_site = set([])
        for se in range(wanted_common_site):
            this_site = find_random_site(input_site, used_site)
            req_sitez.add(this_site)
            if uniqueSE:

                these_SE = SE_to_site_mapping[this_site]
                for this_SE in these_SE:

                    for site in site_to_SE_mapping[this_SE]:
                        used_site.add(site)
            else:
                used_site.add(this_site)
    else:
        for s in input_site:
            req_sitez.add(s)
    return req_sitez


def calculateSiteSEMapping(file_replicas, wanted_common_site, uniqueSE, site_to_SE_mapping, SE_to_site_mapping):

    SE_dict = dict()
    maps_size = 0
    found = []

    # First find the SE for each site
    for lfn, repz in file_replicas.iteritems():
        sitez = set([])
        for replica in repz:
            sitez.add(replica)
            if not replica in found:

                from Ganga.Core.GangaThread.WorkerThreads import getQueues
                getQueues()._monitoring_threadpool.add_function(addToMapping, (str(replica), site_to_SE_mapping))

                maps_size = maps_size + 1
                found.append(replica)

        SE_dict[lfn] = sitez

    # Doing this in parallel so wait for it to finish
    while len(site_to_SE_mapping) != maps_size:
        import time
        time.sleep(0.1)

    # Now calculate the 'inverse' dictionary of site for each SE
    for _SE, _sites in site_to_SE_mapping.iteritems():
        for site_i in _sites:
            if site_i not in SE_to_site_mapping:
                SE_to_site_mapping[site_i] = set([])
            if _SE not in SE_to_site_mapping[site_i]:
                SE_to_site_mapping[site_i].add(_SE)

    # These can be used to select the site which know of a given SE
    # Or vice versa

    # Now lets generate a dictionary of some chosen site vs LFN to use in
    # constructing subsets
    allSubSets = []
    allChosenSets = {}

    site_dict = {}
    for _lfn, sites in SE_dict.iteritems():
        site_dict[_lfn] = set([])
        for _site in sites:
            for _SE in site_to_SE_mapping[_site]:
                site_dict[_lfn].add(_SE)

    # Now select a set of site to use as a seed for constructing a subset of
    # LFN
    for lfn in site_dict.keys():
        allChosenSets[lfn] = generate_site_selection(
            site_dict[lfn], wanted_common_site, uniqueSE, site_to_SE_mapping, SE_to_site_mapping)

    return site_dict, allSubSets, allChosenSets


def lookUpLFNReplicas(inputs, allLFNData):
    # Build a useful dictionary and list
    allLFNs = []
    LFNdict = {}
    for _lfn in inputs:
        allLFNs.append(_lfn.lfn)
        LFNdict[_lfn.lfn] = _lfn

    # Request the replicas for all LFN 'LFN_parallel_limit' at a time to not overload the
    # server and give some feedback as this is going on
    from GangaDirac.Lib.Utilities.DiracUtilities import execute
    import math
    global limit_divide_one
    for i in range(int(math.ceil(float(len(allLFNs)) * limit_divide_one))):

        from Ganga.Core.GangaThread.WorkerThreads import getQueues
        getQueues()._monitoring_threadpool.add_function(getLFNReplicas, (allLFNs, i, allLFNData))

    while len(allLFNData) != int(math.ceil(float(len(allLFNs)) * limit_divide_one)):
        import time
        time.sleep(1.)
        # This can take a while so lets protect any repo locks
        import Ganga.Runtime.Repository_runtime
        Ganga.Runtime.Repository_runtime.updateLocksNow()

    return allLFNs, LFNdict


def sortLFNreplicas(bad_lfns, allLFNs, LFNdict, ignoremissing, allLFNData, inputs):
    from Ganga.GPIDev.Base.Proxy import stripProxy, isType
    from Ganga.GPIDev.Base.Objects import GangaObject

    myRegistry = {}
    for this_LFN in inputs:
        if isType(this_LFN, GangaObject):
            myRegistry[this_LFN.lfn] = stripProxy(this_LFN)._getRegistry()
            if myRegistry[this_LFN.lfn] is not None:
                myRegistry[this_LFN.lfn].turnOffAutoFlushing()

    try:
        return _sortLFNreplicas(bad_lfns, allLFNs, LFNdict, ignoremissing, allLFNData)
    except Exception as err:
        logger.debug("Sorting Exception: %s" % str(err))
        raise err
    finally:
        for this_LFN in myRegistry.keys():
            if myRegistry[this_LFN] is not None:
                myRegistry[this_LFN].turnOnAutoFlushing()

def _sortLFNreplicas(bad_lfns, allLFNs, LFNdict, ignoremissing, allLFNData):

    import math
    from Ganga.GPIDev.Base.Proxy import stripProxy

    errors = []

    # FIXME here to keep the repo settings as they were before we changed the
    # flush count
    original_write_perm = {}

    global LFN_parallel_limit
    global limit_divide_one

    for i in range(int(math.ceil(float(len(allLFNs)) * limit_divide_one))):
        #logger.info( "%s of %s" % (str(i), str(int(math.ceil( float(len(allLFNs))*limit_divide_one )))) )
        output = allLFNData.get(i)

        if output is None:
            logger.error("Error getting Replica information from Dirac: [%s,%s]" % ( str(i * LFN_parallel_limit), str((i + 1) * LFN_parallel_limit)))
            logger.error("%s" % str(allLFNData))
            raise Exception('Error from DIRAC')

        # Identify files which have Failed to be found by DIRAC
        try:
            results = output.get('Value')
            if len(results.get('Failed').keys()) > 0:
                values = results.get('Failed')
                errors.append(str(values))
                #raise SplittingError( "Error getting LFN Replica information:\n%s" % str(values) )
                for this_lfn in results.get('Failed').keys():
                    bad_lfns.append(this_lfn)
        except SplittingError as split_Err:
            raise split_Err
        except Exception as err:
            try:
                error = output
                logger.error("%s" % str(output))
            except Exception as err:
                logger.debug("Error: %s" % str(err))
                pass
            logger.error("Unknown error ion Dirac LFN Failed output")
            raise

        try:
            results = output.get('Value')
            values = results.get('Successful')
        except Exception as err:
            logger.error("Unknown error in parsing Dirac LFN Successful output")
            raise

        upper_limit = (i+1)*LFN_parallel_limit
        if upper_limit > len(allLFNs):
            upper_limit = len(allLFNs)

        logger.debug("Updating LFN Physical Locations: [%s:%s] of %s" % (str(i * LFN_parallel_limit), str(upper_limit), str(len(allLFNs))))

        for this_lfn in values.keys():
            #logger.debug("LFN: %s" % str(this_lfn))
            this_dict = {}
            this_dict[this_lfn] = values.get(this_lfn)

            if this_lfn in LFNdict:
                #logger.debug("Updating RemoteURLs")
                LFNdict[this_lfn]._updateRemoteURLs(this_dict)
                #logger.debug("This_dict: %s" % str(this_dict))
            else:
                logger.error("Error updating remoteURLs for: %s" % str(this_lfn))

            # If we find NO replicas then also mark this as bad!
            if this_dict[this_lfn].keys() == []:
                bad_lfns.append(this_lfn)

        for this_lfn in bad_lfns:
            logger.warning("LFN: %s was either unknown to DIRAC or unavailable, Ganga is ignoring it!" % str(this_lfn))
            if this_lfn in LFNdict:
                del LFNdict[this_lfn]
            if this_lfn in allLFNs:
                allLFNs.remove(this_lfn)

    return errors

# Actually Do the work of the splitting


def OfflineGangaDiracSplitter(_inputs, filesPerJob, maxFiles, ignoremissing):
    """
    Generator that yields a datasets for dirac split jobs
    """

    if maxFiles is not None and maxFiles > 0:
        inputs = _inputs[:maxFiles]
    else:
        inputs = _inputs

    from GangaDirac.Lib.Files.DiracFile import DiracFile
    from Ganga.GPIDev.Adapters.ISplitter import SplittingError
    # First FIND ALL LFN REPLICAS AND SE<->SITE MAPPINGS AND STORE THIS IN MEMORY
    # THIS IS DONE IN PARALLEL TO AVOID OVERLOADING DIRAC WITH THOUSANDS OF
    # REQUESTS AT ONCE ON ONE CONNECTION

    wanted_common_site = configDirac['OfflineSplitterMaxCommonSites']
    iterative_limit = configDirac['OfflineSplitterLimit']
    good_fraction = configDirac['OfflineSplitterFraction']
    uniqueSE = configDirac['OfflineSplitterUniqueSE']

    split_files = []

    if inputs is None:
        raise SplittingError("Cannot Split Job as the inputdata appears to be None!")

    if len(inputs.getLFNs()) != len(inputs.files):
        raise SplittingError("Error trying to split dataset using DIRAC backend with non-DiracFile in the inputdata")

    file_replicas = {}

    logger.info("Requesting LFN replica info")

    allLFNData = {}

    # Perform a lookup of where LFNs are all stored
    allLFNs, LFNdict = lookUpLFNReplicas(inputs, allLFNData)

    for _lfn in allLFNData:
        if allLFNData[_lfn] is None:
            logger.error("Error in Getting LFN Replica information, aborting split")
            raise SplittingError("Error in Getting LFN Replica information, aborting split")

    bad_lfns = []

    # Sort this information and store is in the relevant Ganga objects
    errors = sortLFNreplicas(bad_lfns, allLFNs, LFNdict, ignoremissing, allLFNData, inputs)

    if len(bad_lfns) != 0:
        if ignoremissing is False:
            logger.error("Errors found getting LFNs:\n%s" % str(errors))
            raise SplittingError("Error trying to split dataset with invalid LFN and ignoreMissing = False")

    # This finds all replicas for all LFNs...
    # This will probably struggle for LFNs which don't exist
    # Bad LFN should have been removed by this point however
    all_lfns = [LFNdict[this_lfn].locations for this_lfn in LFNdict if this_lfn not in bad_lfns]

    logger.info("Got replicas")

    for this_input in inputs:
        if this_input.lfn not in bad_lfns:
            file_replicas[this_input.lfn] = this_input.locations

    logger.info("found all replicas")

    logger.info("Calculating site<->SE Mapping")

    site_to_SE_mapping = {}
    SE_to_site_mapping = {}

    # Now lets generate a dictionary of some chosen site vs LFN to use in
    # constructing subsets
    site_dict, allSubSets, allChosenSets = calculateSiteSEMapping(
        file_replicas, wanted_common_site, uniqueSE, site_to_SE_mapping, SE_to_site_mapping)

    logger.debug("Found all SE in use")

    # BELOW IS WHERE THE ACTUAL SPLITTING IS DONE

    logger.info("Calculating best data subsets")

    import math
    iterations = 0
    # Loop over all LFNs
    while len(site_dict.keys()) > 0:

        # LFN left to be used
        # NB: Can't modify this list and iterate over it directly in python
        LFN_instances = site_dict.keys()
        # Already used LFN
        chosen_lfns = []

        for iterating_LFN in LFN_instances:

            # If this has previously been selected lets ignore it and move on
            if iterating_LFN in chosen_lfns:
                continue

            # Use this seed to try and construct a subset
            req_sitez = allChosenSets[iterating_LFN]
            _this_subset = []

            #logger.debug("find common LFN for: " + str(allChosenSets[iterating_LFN]))

            # Construct subset
            # Starting with i, populate subset with LFNs which have an
            # overlap of at least 2 SE
            for this_LFN in LFN_instances:
                if this_LFN in chosen_lfns:
                    continue
                if req_sitez.issubset(site_dict[this_LFN]):
                    if len(_this_subset) >= filesPerJob:
                        break
                    _this_subset.append(this_LFN)

            limit = int(math.floor(float(filesPerJob) * good_fraction))

            #logger.debug("Size limit: %s" % str(limit))

            # If subset is too small throw it away
            if len(_this_subset) < limit:
                #logger.debug("%s < %s" % (str(len(_this_subset)), str(limit)))
                allChosenSets[iterating_LFN] = generate_site_selection(site_dict[iterating_LFN], wanted_common_site, uniqueSE, site_to_SE_mapping, SE_to_site_mapping)
                continue
            else:
                logger.debug("found common LFN for: " + str(allChosenSets[iterating_LFN]))
                logger.debug("%s > %s" % (str(len(_this_subset)), str(limit)))
                # else Dataset was large enough to be considered useful
                logger.debug("Generating Dataset of size: %s" % str(len(_this_subset)))
                ## Construct DiracFile here as we want to keep the above combination
                allSubSets.append([DiracFile(lfn=str(this_LFN)) for this_LFN in _this_subset])

                for lfn in _this_subset:
                    site_dict.pop(lfn)
                    allChosenSets.pop(lfn)
                    chosen_lfns.append(lfn)

        # Lets keep track of how many times we've tried this
        iterations = iterations + 1

        # Can take a while so lets not let threads become un-locked
        import Ganga.Runtime.Repository_runtime
        Ganga.Runtime.Repository_runtime.updateLocksNow()

        # If on final run, will exit loop after this so lets try and cleanup
        if iterations >= iterative_limit:

            if good_fraction < 0.5:
                good_fraction = good_fraction * 0.75
                iterations = 0
            elif wanted_common_site > 1:
                logger.debug("Reducing Common Site Size")
                wanted_common_site = wanted_common_site - 1
                iterations = 0
                good_fraction = 0.75
            else:
                good_fraction = good_fraction * 0.75

            logger.debug("good_fraction: %s" % str(good_fraction))

    split_files = allSubSets

    avg = float()
    for this_set in allSubSets:
        avg += float(len(this_set))
    avg /= float(len(allSubSets))

    logger.info("Average Subset size is: %s" % (str(avg)))

    # FINISHED SPLITTING CHECK!!!

    check_count = 0
    for i in split_files:
        check_count = check_count + len(i)

    if check_count != len(inputs) - len(bad_lfns):
        logger.error("SERIOUS SPLITTING ERROR!!!!!")
        raise SplitterError("Files Missing after Splitting!")
    else:
        logger.info("File count checked! Ready to Submit")

    # RETURN THE RESULT

    logger.info("Created %s subsets" % str(len(split_files)))

    #logger.info( "Split Files: %s" % str(split_files) )

    for dataset in split_files:
        yield dataset




from Ganga.GPIDev.Lib.Tasks.common import makeRegisteredJob, getJobByID
from Ganga.GPIDev.Lib.Tasks.IUnit import IUnit
from Ganga.GPIDev.Lib.Job.Job import JobError
from Ganga.GPIDev.Lib.Registry.JobRegistry import JobRegistrySlice, JobRegistrySliceProxy
from Ganga.Core.exceptions import ApplicationConfigurationError
from Ganga.GPIDev.Base.Proxy import addProxy, stripProxy
from commands import getstatusoutput
from datetime import datetime, date, time

import os

from Ganga.Utility.logging import getLogger
logger = getLogger(modulename=True)

class NA62Unit(IUnit):
   _schema = Schema(Version(1,0), dict(IUnit._schema.datadict.items() + {
    }.items()))

   _category = 'units'
   _name = 'AtlasUnit'
   _exportmethods = IUnit._exportmethods + [ ]

   def __init__(self):
      super(NA62Unit, self).__init__()
      
   def createNewJob(self):
      """Create any jobs required for this unit"""      
      j = makeRegisteredJob()
      j._impl.backend = self._getParent().backend.clone()
      j._impl.application = self._getParent().application.clone()
      j.application.run_number = j.application._impl.getNextRunNumber()
      return j

   def updateStatus(self, status):
      "if we've just been switched to running, the job has been submitted so update the DB"
      if (status == "running") and (self.status == "new") and len(self.active_job_ids) != 0:
         ins_fields = "run,description,decay_type,radcor,mc_version,seed,events,output_name,jdl,mac,exe,stderr,stdout,status_url,submitter,submitted_on,status"
         #(submitted_on format is "2013-MM-DD HH:mm:ss")
         app = getJobByID(self.active_job_ids[0]).application
         ins_vals = "%d, '%s production job', %d, %d, %d, %d, %d, '%s_v%d_r%d.root', '__jdlfile__', '%s', '%s', 'na62run%d.err', 'na62run%d.out', '%s', 'ganga', '%s', 'SUBMITTED'" % (app.run_number, app.decay_name, app.decay_type, app.radcor, app.mc_version, app.run_number, app.num_events, app.file_prefix, app.mc_version, app.run_number, app._impl.getMACFileName(), app.script_name, app.run_number, app.run_number, getJobByID(self.active_job_ids[0]).backend.id, datetime.now().strftime("%Y-%m-%d %H:%m:%S"))

         nec_file = ".gpytho"
         work_dir = "/clusterhome/home/protopop"
         nec_str = open(os.path.join( work_dir, nec_file )).read().strip().strip('#')
         mysqlc = "mysql -hhughnon.ppe.gla.ac.uk -ugridbot -p%s -s GridJobs" % nec_str
      
         rc, out = getstatusoutput("echo \"INSERT INTO jobs (%s) VALUES (%s)\" | %s" % (ins_fields, ins_vals, mysqlc))
         
         if (rc != 0):
             logger.error(out)



from __future__ import absolute_import
from Ganga.Core.GangaRepository import addRegistry
from Ganga.GPIDev.Lib.Tasks.TaskRegistry import TaskRegistry

# Tasks
from Ganga.GPIDev.Lib.Tasks.ITask import ITask
from Ganga.GPIDev.Lib.Tasks.ITransform import ITransform
from Ganga.GPIDev.Lib.Tasks.TaskChainInput import TaskChainInput
from Ganga.GPIDev.Lib.Tasks.TaskLocalCopy import TaskLocalCopy

from Ganga.GPIDev.Lib.Tasks.CoreTask import CoreTask
from Ganga.GPIDev.Lib.Tasks.CoreTransform import CoreTransform
from Ganga.GPIDev.Lib.Tasks.CoreUnit import CoreUnit

# Create the registry
myTaskRegistry = TaskRegistry("tasks", "Tasks Registry")
addRegistry(myTaskRegistry)


def stopTasks():
    global myTaskRegistry
    myTaskRegistry.stop()





from __future__ import absolute_import
##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: LocalFile.py,v 0.1 2011-09-29 15:40:00 idzhunov Exp $
##########################################################################

from Ganga.GPIDev.Schema import Schema, Version, SimpleItem, ComponentItem

from Ganga.GPIDev.Adapters.IGangaFile import IGangaFile

from Ganga.GPIDev.Base.Proxy import GPIProxyObjectFactory

from Ganga.GPIDev.Lib.File.File import File
from Ganga.GPIDev.Lib.File import FileBuffer

import Ganga.Utility.logging

import errno
import re
import os
import os.path
import copy

logger = Ganga.Utility.logging.getLogger()

regex = re.compile('[*?\[\]]')

class LocalFile(IGangaFile):

    """LocalFile represents base class for output files, such as MassStorageFile, LCGSEFile, etc 
    """
    _schema = Schema(Version(1, 1), {'namePattern': SimpleItem(defvalue="", doc='pattern of the file name'),
                                     'localDir': SimpleItem(defvalue="", doc='local dir where the file is stored, used from get and put methods'),
                                     'subfiles': ComponentItem(category='gangafiles', defvalue=[], hidden=1,
                                                sequence=1, copyable=0, doc="collected files from the wildcard namePattern"),
                                     'compressed': SimpleItem(defvalue=False, typelist=[bool], protected=0, doc='wheather the output file should be compressed before sending somewhere'),
                                     #'output_location': SimpleItem(defvalue=None, typelist=[str, None], hidden=1, copyable=1, doc="path of output location on disk")
                                     })
    _category = 'gangafiles'
    _name = "LocalFile"
    _exportmethods = ["location", "remove", "accessURL", "processOutputWildcardMatches", "hasMatchedFiles"]

    def __init__(self, namePattern='', localDir='', **kwds):
        """ name is the name of the output file that is going to be processed
            in some way defined by the derived class
        """
        super(LocalFile, self).__init__()

        self.tmp_pwd = None
        self.output_location = None

        if isinstance(namePattern, str):
            self.namePattern = namePattern
        elif isinstance(namePattern, File):
            self.namePattern = os.path.basename(namePattern.name)
            self.localDir = os.path.dirname(namePattern.name)
        elif isinstance(namePattern, FileBuffer):
            namePattern.create()
            self.namePattern = os.path.basename(namePattern.name)
            self.localDir = os.path.dirname(namePattern.name)
        else:
            logger.error("Unkown type: %s . Cannot Create LocalFile from this!" % type(namePattern))

        if isinstance(localDir, str):
            if localDir != '':
                self.localDir = localDir
            else:
                this_pwd = os.path.abspath('.')
                self.tmp_pwd = this_pwd
        else:
            logger.error("Unkown type: %s . Cannot set LocalFile localDir using this!" % type(localDir))

    def __construct__(self, args):

        self.tmp_pwd = None
        self.output_location = None

        self.localDir = ''

        from Ganga.GPIDev.Lib.File.SandboxFile import SandboxFile
        if len(args) == 1 and isinstance(args[0], str):
            self.namePattern = args[0]
        elif len(args) == 2 and isinstance(args[0], str) and isinstance(args[1], str):
            self.namePattern = args[0]
            self.localDir = args[1]
        elif len(args) == 1 and isinstance(args[0], SandboxFile):
            super(LocalFile, self).__construct__(args)

        if self.localDir == '' and self.namePattern != '':
            this_pwd = os.path.abspath('.')
            self.tmp_pwd = this_pwd

    def __repr__(self):
        """Get the representation of the file."""

        return "LocalFile(namePattern='%s')" % self.namePattern

    def location(self):
        return self.getFilenameList()

    def accessURL(self):
        URLs = []
        for file in self.location():
            URLs.append('file://' + os.path.join(os.sep, file))
        return URLs

    def processOutputWildcardMatches(self):
        """This collects the subfiles for wildcarded output LocalFile"""
        import glob

        fileName = self.namePattern

        if self.compressed:
            fileName = '%s.gz' % self.namePattern

        sourceDir = self.getJobObject().outputdir
        if regex.search(fileName) is not None:

            for currentFile in glob.glob(os.path.join(sourceDir, fileName)):

                d = LocalFile(namePattern=os.path.basename(currentFile))
                d.compressed = self.compressed

                self.subfiles.append(GPIProxyObjectFactory(d))

    def processWildcardMatches(self):

        if self.subfiles:
            return self.subfiles

        import glob

        fileName = self.namePattern

        if self.compressed:
            fileName = '%s.gz' % self.namePattern

        sourceDir = self.localDir

        if regex.search(fileName) is not None:
            for currentFile in glob.glob(os.path.join(sourceDir, fileName)):
                d = LocalFile(namePattern=os.path.basename(
                    currentFile), localDir=os.path.dirname(currentFile))
                d.compressed = self.compressed

                self.subfiles.append(GPIProxyObjectFactory(d))

    def getFilenameList(self):
        """Return the files referenced by this LocalFile"""
        filelist = []
        self.processWildcardMatches()
        if self.subfiles:
            for f in self.subfiles:
                filelist.append(os.path.join(f.localDir, f.namePattern))
        else:
            if self.localDir == '':
                if os.path.exists(os.path.join(self.tmp_pwd, self.namePattern)):
                    self.localDir = self.tmp_pwd
                    logger.debug("File: %s found, Setting localDir: %s" % (self.namePattern, self.localDir))
                else:
                    this_pwd = os.path.abspath('.')
                    now_tmp_pwd = this_pwd
                    if os.path.exists(os.path.join(now_tmp_pwd, self.namePattern)):
                        self.localDir = now_tmp_pwd
                        logger.debug("File: %s found, Setting localDir: %s" % (self.namePattern, self.localDir))
                    else:
                        logger.debug("File: %s NOT found, NOT setting localDir: %s !!!" % (self.namePattern, self.localDir))

            filelist.append(os.path.join(self.localDir, self.namePattern))

        return filelist

    def hasMatchedFiles(self):
        """
        OK for checking subfiles but of no wildcards, need to actually check file exists
        """

        # check for subfiles
        if len(self.subfiles) > 0:
            # we have subfiles so we must have actual files associated
            return True
        else:
            if self.containsWildcards():
                return False

        # check if single file exists (no locations field to try)
        job = self.getJobObject()
        if job:
            fname = self.namePattern
            if self.compressed:
                fname += ".gz"

            if os.path.isfile(os.path.join(job.getOutputWorkspace().getPath(), fname)):
                return True

        return False

    def remove(self):

        for this_file in self.getFilenameList():
            _actual_delete = False
            keyin = None
            while keyin is None:
                keyin = raw_input("Do you want to remove the LocalFile: %s ? ([y]/n) " % this_file)
                if keyin in ['y', '']:
                    _actual_delete = True
                elif keyin == 'n':
                    _actual_delete = False
                else:
                    logger.warning("y/n please!")
                    keyin = None
            if _actual_delete:
                if not os.path.exists(this_file):
                    logger.warning(
                        "File %s did not exist, can't delete" % this_file)
                else:
                    logger.info("Deleting: %s" % this_file)

                    import time
                    remove_filename = this_file + "_" + str(time.time()) + '__to_be_deleted_'
                    try:
                        os.rename(this_file, remove_filename)
                    except Exception as err:
                        logger.warning("Error in first stage of removing file: %s" % this_file)
                        remove_filename = this_file

                    try:
                        os.remove(remove_filename)
                    except OSError as err:
                        if err.errno != errno.ENOENT:
                            logger.error("Error in removing file: %s" % remove_filename)
                            raise
                        pass

        return


## rcurrie Attempted to implement for 6.1.9 but commenting out due to not being able to correctly make use of setLocation

#    def getWNScriptDownloadCommand(self, indent):
#
#        script = """
####INDENT###os.system('###CP_COMMAND')
#
#"""
#        full_path = os.path.join(self.localDir, self.namePattern)
#        replace_dict = {'###INDENT###' : indent, '###CP_COMMAND###' : 'cp %s .' % full_path}
#
#        for k, v in replace_dict.iteritems():
#            script = script.replace(k, v)
#
#        return script
#
#
#    def getWNInjectedScript(self, outputFiles, indent, patternsToZip, postProcessLocationsFP):
#
#        cp_template = """
####INDENT###os.system("###CP_COMMAND###")
#"""
#        script = ""
#
#        for this_file in outputFiles:
#            filename = this_file.namePattern
#            cp_cmd = 'cp %s %s' % (filename, self.output_location)
#
#            this_cp = str(cp_template)
#
#            replace_dict = {'###INDENT###' : indent, '###CP_COMMAND###' : cp_cmd}
#
#            for k, v in replace_dict.iteritems():
#                print("Replace %s : %s" % (k, v))
#                this_cp = this_cp.replace(k, v)
#
#            script = this_cp
#            break
#
#        return script
#
#    def setLocation(self):
#
#        job = self.getJobObject()
#
#        self.output_location = job.getOutputWorkspace(create=True).getPath()
#
#        return




#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#
"""The Ganga backendhandler for the Dirac system."""

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#
import os
import re
import fnmatch
import time
from Ganga.GPIDev.Schema import Schema, Version, SimpleItem
from Ganga.GPIDev.Adapters.IBackend import IBackend
from Ganga.Core import BackendError, GangaException
from GangaDirac.Lib.Backends.DiracUtils import result_ok, get_job_ident, get_parametric_datasets, outputfiles_iterator, outputfiles_foreach
from GangaDirac.Lib.Files.DiracFile import DiracFile
from GangaDirac.Lib.Utilities.DiracUtilities import execute, _proxyValid
from Ganga.Utility.ColourText import getColour
from Ganga.Utility.Config import getConfig
from Ganga.Utility.logging import getLogger
from Ganga.GPIDev.Credentials import getCredential
from Ganga.GPIDev.Base.Proxy import stripProxy, isType, getName
logger = getLogger()
regex = re.compile('[*?\[\]]')


class DiracBase(IBackend):

    """The backend that submits jobs to the Grid via DIRAC.

    The backend for jobs to be submitted to the Grid. Jobs are
    submitted through the DIRAC WMS system and then in turn submitted to the
    Grid. A few examples of usage are given below

    # Create Dirac backend object
    b = Dirac()

    # Create and submit job.
    j = Job(application=app,backend=b)
    j.submit()

    # Run a Root job on the Grid if in LHCb VO

    # Create a Root application object. See Root help text for instructions
    # on how to configure this.
    app = Root()

    # Create and submit job to Dirac using default options
    j = Job(application=app,backend=Dirac())
    j.submit()

    # Using the 'settings' attribute
    j.backend.settings['BannedSites'] = ['LCG.CERN.ch']
    j.resubmit()

    # settings can be set at any time but are only 'respected' during
    # submit and resubmit.

    """

    dirac_monitoring_is_active = True

    _schema = Schema(Version(3, 2), {
        'id': SimpleItem(defvalue=None, protected=1, copyable=0,
                         typelist=['int', 'type(None)'],
                         doc='The id number assigned to the job by the DIRAC WMS. If seeking help'
                         ' on jobs with the Dirac backend, please always report this id '
                         'number in addition to a full description of your problem. The id '
                         'can also be used to further inspect the job at '
                         'https://lhcbweb.pic.es/DIRAC/info/general/diracOverview'),
        'status': SimpleItem(defvalue=None, protected=1, copyable=0,
                             typelist=['str', 'type(None)'],
                             doc='The detailed status as reported by the DIRAC WMS'),
        'actualCE': SimpleItem(defvalue=None, protected=1, copyable=0,
                               typelist=['str', 'type(None)'],
                               doc='The location where the job ran'),
        'normCPUTime': SimpleItem(defvalue=None, protected=1, copyable=0,
                                  typelist=['str', 'type(None)'],
                                  doc='The normalized CPU time reported by the DIRAC WMS'),
        'statusInfo': SimpleItem(defvalue='', protected=1, copyable=0,
                                 typelist=['str', 'type(None)'],
                                 doc='Minor status information from Dirac'),
        'extraInfo': SimpleItem(defvalue='', protected=1, copyable=0,
                                typelist=['str', 'type(None)'],
                                doc='Application status information from Dirac'),
        'diracOpts': SimpleItem(defvalue='',
                                doc='DIRAC API commands to add the job definition script. Only edit '
                                'if you *really* know what you are doing'),
        'settings': SimpleItem(defvalue={'CPUTime': 2 * 86400},
                               doc='Settings for DIRAC job (e.g. CPUTime, BannedSites, etc.)')
    })
    _exportmethods = ['getOutputData', 'getOutputSandbox', 'removeOutputData',
                      'getOutputDataLFNs', 'peek', 'reset', 'debug']
    _packed_input_sandbox = True
    _category = "backends"
    _name = 'DiracBase'
    _hidden = True

    def _setup_subjob_dataset(self, dataset):
        return None

    def _setup_bulk_subjobs(self, dirac_ids, dirac_script):
        f = open(dirac_script, 'r')
        parametric_datasets = get_parametric_datasets(f.read().split('\n'))
        f.close()
        if len(parametric_datasets) != len(dirac_ids):
            raise BackendError('Dirac', 'Missmatch between number of datasets defines in dirac API script and those returned by DIRAC')

        from Ganga.GPIDev.Lib.Job.Job import Job
        master_job = self.getJobObject()
        master_job.subjobs = []
        for i in range(len(dirac_ids)):
            j = Job()
            j.copyFrom(master_job)
            j.splitter = None
            j.backend.id = dirac_ids[i]
            j.id = i
            j.inputdata = self._setup_subjob_dataset(parametric_datasets[i])
            j.status = 'submitted'
            j.time.timenow('submitted')
            master_job.subjobs.append(j)
        master_job._commit()
        return True

    def _common_submit(self, dirac_script):
        '''Submit the job via the Dirac server.'''
        j = self.getJobObject()
        self.id = None
        self.actualCE = None
        self.status = None
        self.extraInfo = None
        self.statusInfo = ''
        j.been_queued = False
        dirac_cmd = """execfile(\'%s\')""" % dirac_script
        result = execute(dirac_cmd)
        # Could use the below code instead to submit on a thread
        # If submitting many then user may terminate ganga before
        # all jobs submitted
#        def submit_checker(result, job, script):
#            err_msg = 'Error submitting job to Dirac: %s' % str(result)
#            if not result_ok(result) or 'Value' not in result:
#                logger.error(err_msg)
#                raise BackendError('Dirac',err_msg)
#
#            idlist = result['Value']
#            if type(idlist) is list:
#                return job._setup_bulk_subjobs(idlist, script)
#            job.id = idlist
#        server.execute_nonblocking(dirac_cmd, callback_func=submit_checker, args=(self, dirac_script))
#        return True

        err_msg = 'Error submitting job to Dirac: %s' % str(result)
        if not result_ok(result) or 'Value' not in result:
            logger.error(err_msg)
            logger.error("\n\n===\n%s\n===\n" % dirac_script)
            logger.error("\n\n====\n")
            with open(dirac_script, 'r') as file_in:
                logger.error("%s" % file_in.read())
            logger.error("\n====\n")
            raise BackendError('Dirac', err_msg)

        idlist = result['Value']
        if type(idlist) is list:
            return self._setup_bulk_subjobs(idlist, dirac_script)

        self.id = idlist
        return type(self.id) == int

    def _addition_sandbox_content(self, subjobconfig):
        '''any additional files that should be sent to dirac'''
        return []

    def submit(self, subjobconfig, master_input_sandbox):
        """Submit a DIRAC job"""
        j = self.getJobObject()

        sboxname = j.createPackedInputSandbox(subjobconfig.getSandboxFiles())

        input_sandbox = master_input_sandbox[:]
        input_sandbox += sboxname

        input_sandbox += self._addition_sandbox_content(subjobconfig)

        ## Add LFN to the inputfiles section of the file
        input_sandbox_userFiles = []
        for this_file in j.inputfiles:
            if isType(this_file, DiracFile):
                input_sandbox_userFiles.append('LFN:'+str(this_file.lfn))
        if j.master:
            for this_file in j.master.inputfiles:
                if isType(this_file, DiracFile):
                    input_sandbox_userFiles.append('LFN:'+str(this_file.lfn))

        for this_file in input_sandbox_userFiles:
            input_sandbox.append(this_file)

        dirac_script = subjobconfig.getExeString().replace('##INPUT_SANDBOX##', str(input_sandbox))

        dirac_script_filename = os.path.join(j.getInputWorkspace().getPath(), 'dirac-script.py')
        f = open(dirac_script_filename, 'w')
        f.write(dirac_script)
        f.close()
        return self._common_submit(dirac_script_filename)

    def master_auto_resubmit(self, rjobs):
        '''Duplicate of the IBackend.master_resubmit but hooked into auto resubmission
        such that the monitoring server is used rather than the user server'''
        from Ganga.Core import IncompleteJobSubmissionError, GangaException
        from Ganga.Utility.logging import log_user_exception
        incomplete = 0

        def handleError(x):
            if incomplete:
                raise x
            else:
                return 0
        try:
            for sj in rjobs:
                fqid = sj.getFQID('.')
                logger.info("resubmitting job %s to %s backend", fqid, getName(sj.backend))
                try:
                    b = sj.backend
                    sj.updateStatus('submitting')
                    result = b._resubmit()
                    if result:
                        sj.updateStatus('submitted')
                        # sj._commit() # PENDING: TEMPORARY DISABLED
                        incomplete = 1
                    else:
                        return handleError(IncompleteJobSubmissionError(fqid, 'resubmission failed'))
                except Exception as x:
                    log_user_exception(logger, debug=isType(x, GangaException))
                    return handleError(IncompleteJobSubmissionError(fqid, str(x)))
        finally:
            master = self.getJobObject().master
            if master:
                master.updateMasterJobStatus()
        return 1

    def resubmit(self):
        """Resubmit a DIRAC job"""
        return self._resubmit()

    def _resubmit(self):
        """Resubmit a DIRAC job"""
        j = self.getJobObject()
        parametric = False
        script_path = os.path.join(j.getInputWorkspace().getPath(), 'dirac-script.py')
        # Check old script
        if j.master is None and not os.path.exists(script_path):
            raise BackendError('Dirac', 'No "dirac-script.py" found in j.inputdir')

        if j.master is not None and not os.path.exists(script_path):
            script_path = os.path.join(
                j.master.getInputWorkspace().getPath(), 'dirac-script.py')
            if not os.path.exists(script_path):
                raise BackendError('Dirac', 'No "dirac-script.py" found in j.inputdir or j.master.inputdir')
            parametric = True

        # Read old script
        f = open(script_path, 'r')
        script = f.read()
        f.close()

        # Create new script - ##note instead of using get_parametric_dataset
        # could just use j.inputdata.
        if parametric is True:
            parametric_datasets = get_parametric_datasets(script.split('\n'))
            if j.master:
                if len(parametric_datasets) != len(j.master.subjobs):
                    raise BackendError('Dirac', 'number of parametric datasets defined in API script doesn\'t match number of master.subjobs')
            if j.inputdata and len(j.inputdata) > 0:
                _input_files = [f for f in j.inputdata if not isType(f, DiracFile)]
            else:
                _input_files = []
            if set(parametric_datasets[j.id]).symmetric_difference(set([f.namePattern for f in _input_files])):
                raise BackendError(
                    'Dirac', 'Mismatch between dirac-script and job attributes.')
            script = script.replace('.setParametricInputData(%s)' % str(parametric_datasets),
                                    '.setInputData(%s)' % str(parametric_datasets[j.id]))
            script = script.replace('%n', str(j.id))  # name

        start_user_settings = '# <-- user settings\n'
        new_script = script[
            :script.find(start_user_settings) + len(start_user_settings)]

        job_ident = get_job_ident(script.split('\n'))
        for key, value in self.settings.iteritems():
            if str(key).startswith('set'):
                _key = key[3:]
            else:
                _key = key
            if type(value) is str:
                template = '%s.set%s("%s")\n'
            else:
                template = '%s.set%s(%s)\n'
            new_script += template % (job_ident, str(_key), str(value))
        new_script += script[script.find('# user settings -->'):]

        # Save new script
        new_script_filename = os.path.join(j.getInputWorkspace().getPath(), 'dirac-script.py')
        f = open(new_script_filename, 'w')
        f.write(new_script)
        f.flush()
        f.close()
        return self._common_submit(new_script_filename)

    def reset(self, doSubjobs=False):
        """Resets the state of a job back to 'submitted' so that the
        monitoring will run on it again."""
        j = self.getJobObject()

        disallowed = ['submitting', 'killed']
        if j.status in disallowed:
            logger.warning("Can not reset a job in status '%s'." % j.status)
        else:
            j.getOutputWorkspace().remove(preserve_top=True)
            self.extraInfo = None
            self.statusInfo = ''
            self.status = None
            self.actualCE = None
            j.been_queued = False
            j.updateStatus('submitted')
            if j.subjobs and not doSubjobs:
                logger.info('This job has subjobs, if you would like the backends '
                            'of all the subjobs that are in status=\'completing\' or '
                            'status=\'failed\' also reset then recall reset with the '
                            'arg \'True\' i.e. job(3).backend.reset(True)')
            elif j.subjobs and doSubjobs:
                logger.info('resetting the backends of \'completing\' and \'failed\' subjobs.')
                for sj in j.subjobs:
                    if sj.status == 'completing' or sj.status == 'failed':
                        sj.backend.reset()
            if j.master:
                j.master.updateMasterJobStatus()

    def kill(self):
        """ Kill a Dirac jobs"""
        if not self.id:
            return None
        dirac_cmd = 'kill(%d)' % self.id
        result = execute(dirac_cmd)
        if not result_ok(result):
            raise BackendError('Dirac', 'Could not kill job: %s' % str(result))
        return result['OK']

    def peek(self, filename=None, command=None):
        """Peek at the output of a job (Note: filename/command are ignored)."""
        dirac_cmd = 'peek(%d)' % self.id
        result = execute(dirac_cmd)
        if result_ok(result):
            logger.info(result['Value'])
        else:
            logger.error("No peeking available for Dirac job '%i'.", self.id)

    def getOutputSandbox(self, outputDir=None):
        j = self.getJobObject()
        if outputDir is None:
            outputDir = j.getOutputWorkspace().getPath()
        dirac_cmd = "getOutputSandbox(%d,'%s')"  % (self.id, outputDir)
        result = execute(dirac_cmd)
        if not result_ok(result):
            msg = 'Problem retrieving output: %s' % str(result)
            logger.warning(msg)
            return False

        return True

    def removeOutputData(self):
        """
        Remove all the LFNs associated with this job.
        """
        # Note when the API can accept a list for removeFile I will change
        # this.
        j = self.getJobObject()
        if j.subjobs:
            for sj in j.subjobs:
                outputfiles_foreach(sj, DiracFile, lambda x: x.remove())
        else:
            outputfiles_foreach(j, DiracFile, lambda x: x.remove())

    def getOutputData(self, outputDir=None, names=None, force=False):
        """Retrieve data stored on SE to dir (default=job output workspace).
        If names=None, then all outputdata is downloaded otherwise names should
        be a list of files to download. If force=True then data will be redownloaded
        even if the file already exists.

        Note that if called on a master job then all subjobs outputwill be downloaded.
        If dir is None then the subjobs output goes into their individual
        outputworkspaces as expected. If however one specifies a dir then this is
        treated as a top dir and a subdir for each job will be created below it. This
        will avoid overwriting files with the same name from each subjob.
        """
        j = self.getJobObject()
        if outputDir is not None and not os.path.isdir(outputDir):
            raise GangaException("Designated outupt path '%s' must exist and be a directory" % outputDir)

        def download(dirac_file, job, is_subjob=False):
            dirac_file.localDir = job.getOutputWorkspace().getPath()
            if outputDir is not None:
                output_dir = outputDir
                if is_subjob:
                    output_dir = os.path.join(outputDir, job.fqid)
                    if not os.path.isdir(output_dir):
                        os.mkdir(output_dir)
                dirac_file.localDir = output_dir
            if os.path.exists(os.path.join(dirac_file.localDir, os.path.basename(dirac_file.lfn))) and not force:
                return
            try:
                if isType(dirac_file, DiracFile):
                    dirac_file.get(localPath=dirac_file.localDir)
                else:
                    dirac_file.get()
                return dirac_file.lfn
            # should really make the get method throw if doesn't suceed. todo
            except GangaException as e:
                logger.warning(e)

        suceeded = []
        if j.subjobs:
            for sj in j.subjobs:
                suceeded.extend([download(f, sj, True) for f in outputfiles_iterator(sj, DiracFile) if f.lfn != '' and (names is None or f.namePattern in names)])
        else:
            suceeded.extend([download(f, j, False) for f in outputfiles_iterator(j, DiracFile) if f.lfn != '' and (names is None or f.namePattern in names)])

        return filter(lambda x: x is not None, suceeded)

    def getOutputDataLFNs(self):
        """Retrieve the list of LFNs assigned to outputdata"""
        j = self.getJobObject()
        lfns = []

        if j.subjobs:
            for sj in j.subjobs:
                lfns.extend([f.lfn for f in outputfiles_iterator(sj, DiracFile) if f.lfn != ''])
        else:
            lfns.extend([f.lfn for f in outputfiles_iterator(j, DiracFile) if f.lfn != ''])
        return lfns

    def debug(self):
        '''Obtains some (possibly) useful DIRAC debug info. '''
        # check services
        cmd = 'getServicePorts()'
        result = execute(cmd)
        if type(result) == str:
            import datetime
            try:
                result = eval(result)
            except Exception as err:
                logger.debug("Exception, err: %s" % str(err))
                pass
        if not result_ok(result):
            logger.warning('Could not obtain services: %s' % str(result))
            return
        services = result.get('Value', {})
        for category in services:
            system, service = category.split('/')
            cmd = "ping('%s','%s')" % (system, service)
            result = execute(cmd)
            if type(result) == str:
                import datetime
                try:
                    result = eval(result)
                except Exception as err:
                    logger.debug("Exception: %s" % str(err))
                    pass
            msg = 'OK.'
            if not result_ok(result):
                msg = '%s' % result['Message']
            logger.info('%s: %s' % (category, msg))
        # get pilot info for this job
        if type(self.id) != int:
            return
        j = self.getJobObject()
        cwd = os.getcwd()
        debug_dir = j.getDebugWorkspace().getPath()
        cmd = "getJobPilotOutput(%d,'%s')" % \
              (self.id, debug_dir)
        result = execute(cmd)
        if result_ok(result):
            logger.info('Pilot Info: %s/pilot_%d/std.out.' %
                        (debug_dir, self.id))
        else:
            logger.error(result.get('Message', ''))

    @staticmethod
    def _getStateTime(job, status):
        """Returns the timestamps for 'running' or 'completed' by extracting
        their equivalent timestamps from the loggingInfo."""
        # Now private to stop server cross-talk from user thread. Since updateStatus calles
        # this method whether called itself by the user thread or monitoring thread.
        # Now don't use hook but define our own private version
        # used in monitoring loop... messy but works.
        if job.status != status:
            b_list = ['running', 'completing', 'completed', 'failed']
            backend_final = ['failed', 'completed']
            # backend stamps
            if not job.subjobs and status in b_list:
                for childstatus in b_list:
                    if job.backend.id:
                        logger.debug("Accessing getStateTime() in diracAPI")
                        dirac_cmd = "getStateTime(%d,\'%s\')" % (job.backend.id, childstatus)
                        be_statetime = execute(dirac_cmd)
                        if childstatus in backend_final:
                            job.time.timestamps["backend_final"] = be_statetime
                            logger.debug("Wrote 'backend_final' to timestamps.")
                        else:
                            job.time.timestamps["backend_" + childstatus] = be_statetime
                            logger.debug("Wrote 'backend_%s' to timestamps.", childstatus)
                    if childstatus == status:
                        break
            logger.debug("_getStateTime(job with id: %d, '%s') called.", job.id, job.status)
        else:
            logger.debug("Status changed from '%s' to '%s'. No new timestamp was written", job.status, status)

    def timedetails(self):
        """Prints contents of the loggingInfo from the Dirac API."""
        if not self.id:
            return None
        logger.debug("Accessing timedetails() in diracAPI")
        dirac_cmd = 'timedetails(%d)' % self.id
        return execute(dirac_cmd)

    def job_finalisation_cleanup(job, updated_dirac_status):

        logger = getLogger()

        #   Revert job back to running state if we exit uncleanly
        if job.status == "completing":
            job.updateStatus('running')
            if job.master:
                job.master.updateMasterJobStatus()
        # FIXME should I add something here to cleanup on sandboxes pulled from
        # malformed job output?

    @staticmethod
    def _internal_job_finalisation(job, updated_dirac_status):

        logger = getLogger()

        if updated_dirac_status == 'completed':
            # firstly update job to completing
            DiracBase._getStateTime(job, 'completing')
            if job.status in ['removed', 'killed']:
                return
            if (job.master and job.master.status in ['removed', 'killed']):
                return  # user changed it under us

            job.updateStatus('completing')
            if job.master:
                job.master.updateMasterJobStatus()

            import time
            start = time.time()
            # contact dirac for information
            job.backend.normCPUTime = execute('normCPUTime(%d)' % job.backend.id)
            getSandboxResult = execute("getOutputSandbox(%d,'%s')" % (job.backend.id, job.getOutputWorkspace().getPath()))
            file_info_dict = execute('getOutputDataInfo(%d)' % job.backend.id)
            now = time.time()
            logger.debug('Job ' + job.fqid + ' Time for Dirac metadata : ' + str(now - start))

            logger.debug('Job ' + job.fqid + ' OutputDataInfo: ' + str(file_info_dict))
            logger.debug('Job ' + job.fqid + ' OutputSandbox: ' + str(getSandboxResult))

            # Set DiracFile metadata
            wildcards = [f.namePattern for f in job.outputfiles.get(DiracFile) if regex.search(f.namePattern) is not None]

            with open(os.path.join(job.getOutputWorkspace().getPath(), getConfig('Output')['PostProcessLocationsFileName']), 'ab') as postprocesslocationsfile:
                if not hasattr(file_info_dict, 'keys'):
                    logger.error("Error understanding OutputDataInfo: %s" % str(file_info_dict))
                    from Ganga.Core.exceptions import GangaException
                    raise GangaException("Error understanding OutputDataInfo: %s" % str(file_info_dict))

                for file_name in file_info_dict.get('Value', []):
                    file_name = os.path.basename(file_name)
                    info = file_info_dict.get(file_name)
                    logger.debug("file_name: %s,\tinfo: %s" % (str(file_name), str(info)))

                    valid_wildcards = [wc for wc in wildcards if fnmatch.fnmatch(file_name, wc)]
                    if len(valid_wildcards) == 0: valid_wildcards.append('')

                    if not hasattr(info, 'get'):
                        logger.error("Error getting OutputDataInfo for: %s" % str(job.getFQID('.')))
                        logger.error("Please check the Dirac Job still exists or attempt a job.backend.reset() to try again!")
                        logger.error("Err: %s" % str(info))
                        logger.error("file_info_dict: %s" % str(file_info_dict))
                        from Ganga.Core.exceptions import GangaException
                        raise GangaException("Error getting OutputDataInfo")

                    for wc in valid_wildcards:
                        logger.debug("wildcard: %s" % str(wc))

                        DiracFileData = 'DiracFile:::%s&&%s->%s:::%s:::%s\n' % (wc,
                                                                                file_name,
                                                                                info.get('LFN', 'Error Getting LFN!'),
                                                                                str(info.get('LOCATIONS', ['NotAvailable'])),
                                                                                info.get('GUID', 'NotAvailable')
                                                                                )
                        logger.debug("DiracFileData: %s" % str(DiracFileData))
                        postprocesslocationsfile.write(DiracFileData)

            # check outputsandbox downloaded correctly
            if not result_ok(getSandboxResult):
                logger.warning('Problem retrieving outputsandbox: %s' % str(getSandboxResult))
                DiracBase._getStateTime(job, 'failed')
                if job.status in ['removed', 'killed']:
                    return
                if (job.master and job.master.status in ['removed', 'killed']):
                    return  # user changed it under us
                job.updateStatus('failed')
                if job.master:
                    job.master.updateMasterJobStatus()
                raise BackendError('Problem retrieving outputsandbox: %s' % str(getSandboxResult))

            # finally update job to completed
            DiracBase._getStateTime(job, 'completed')
            if job.status in ['removed', 'killed']:
                return
            if (job.master and job.master.status in ['removed', 'killed']):
                return  # user changed it under us
            job.updateStatus('completed')
            if job.master:
                job.master.updateMasterJobStatus()
            now = time.time()
            logger.debug('Job ' + job.fqid + ' Time for complete update : ' + str(now - start))

        elif updated_dirac_status == 'failed':
            # firstly update status to failed
            DiracBase._getStateTime(job, 'failed')
            if job.status in ['removed', 'killed']:
                return
            if (job.master and job.master.status in ['removed', 'killed']):
                return  # user changed it under us
            job.updateStatus('failed')
            if job.master:
                job.master.updateMasterJobStatus()

            # if requested try downloading outputsandbox anyway
            if getConfig('DIRAC')['failed_sandbox_download']:
                execute("getOutputSandbox(%d,'%s')" %
                        (job.backend.id, job.getOutputWorkspace().getPath()))
        else:
            logger.error("Unexpected dirac status '%s' encountered" % updated_dirac_status)

    @staticmethod
    def job_finalisation(job, updated_dirac_status):

        count = 1
        limit = 5
        sleep_length = 2.5

        while count != limit:

            try:
                count += 1
                if job.status != "running" and (not job.status in ['completed', 'killed', 'removed']):
                    job.updateStatus('submitted')
                    job.updateStatus('running')
                if job.status in ['completed', 'killed', 'removed']:
                    break
                DiracBase._internal_job_finalisation(job, updated_dirac_status)
                break
            except Exception as err:

                logger.warning("An error occured finalising job: %s" % job.getFQID('.'))
                logger.warning("Attemting again (%s of %s) after %s-sec delay" % (str(count), str(limit), str(sleep_length)))
                if count == limit:
                    logger.error("Unable to finalise job after %s retries due to error:\n%s" % (job.getFQID('.'), str(err)))
                    job.force_status('failed')
                    raise err

            time.sleep(sleep_length)

    @staticmethod
    def updateMonitoringInformation(_jobs):
        """Check the status of jobs and retrieve output sandboxes"""
        # Only those jobs in 'submitted','running' are passed in here for checking
        # if however they have already completed in Dirac they may have been put on queue
        # for processing from last time. These should be put back on queue without
        # querying dirac again. Their signature is status = running and job.backend.status
        # already set to Done or Failed etc.

        jobs = [stripProxy(j) for j in _jobs]

        logger = getLogger()

        # make sure proxy is valid
        if not _proxyValid():
            if DiracBase.dirac_monitoring_is_active:
                logger.warning('DIRAC monitoring inactive (no valid proxy found).')
                DiracBase.dirac_monitoring_is_active = False
            return
        else:
            DiracBase.dirac_monitoring_is_active = True

        # remove from consideration any jobs already in the queue. Checking this non persisted attribute
        # is better than querying the queue as cant tell if a job has just been taken off queue and is being processed
        # also by not being persistent, this attribute automatically allows queued jobs from last session to be considered
        # for requeing
        interesting_jobs = [j for j in jobs if not j.been_queued]
        # status that correspond to a ganga 'completed' or 'failed' (see DiracCommands.status(id))
        # if backend status is these then the job should be on the queue
        requeue_dirac_status = {'Completed': 'completed',
                                'Done': 'completed',
                                'Failed': 'failed',
                                'Deleted': 'failed',
                                'Unknown: No status for Job': 'failed'}

        monitor_jobs = [j for j in interesting_jobs if j.backend.status not in requeue_dirac_status]
        requeue_jobs = [j for j in interesting_jobs if j.backend.status in requeue_dirac_status]

        logger.debug('Interesting jobs: ' + repr([j.fqid for j in interesting_jobs]))
        logger.debug('Monitor jobs    : ' + repr([j.fqid for j in monitor_jobs]))
        logger.debug('Requeue jobs    : ' + repr([j.fqid for j in requeue_jobs]))

        from Ganga.Core.GangaThread.WorkerThreads import getQueues

        # requeue existing completed job
        for j in requeue_jobs:
            #            if j.backend.status in requeue_dirac_status:
            getQueues()._monitoring_threadpool.add_function(DiracBase.job_finalisation,
                                                       args=(j, requeue_dirac_status[j.backend.status]),
                                                       priority=5, name="Job %s Finalizing" % j.fqid)
            j.been_queued = True

        # now that can submit in non_blocking mode, can see jobs in submitting
        # that have yet to be assigned an id so ignore them
        # NOT SURE THIS IS VALID NOW BULK SUBMISSION IS GONE
        # EVEN THOUGH COULD ADD queues.add(j.submit) WILL KEEP AN EYE ON IT
#        dirac_job_ids    = [ j.backend.id for j in monitor_jobs if j.backend.id is not None ]
        # Correction this did become a problem for a crashed session during
        # submit, see #104454
        dead_jobs = (j for j in monitor_jobs if j.backend.id is None)
        for d in dead_jobs:
            d.updateStatus('failed')
            if d.master is not None:
                d.master.updateMasterJobStatus()

        ganga_job_status = [j.status for j in monitor_jobs if j.backend.id is not None]
        dirac_job_ids = [j.backend.id for j in monitor_jobs if j.backend.id is not None]

        #logger.debug("GangaStatus: %s" % str(ganga_job_status))
        #logger.debug("diracJobIDs: %s" % str(dirac_job_ids))

        result = execute('status(%s)' % str(dirac_job_ids))

        if len(result) != len(ganga_job_status):
            logger.warning('Dirac monitoring failed fro %s, result = %s' % (
                str(dirac_job_ids), str(result)))
            return

        #logger.debug("%s, %s, %s" % (str(len(ganga_job_status)), str(len(dirac_job_ids)), str(len(result))))

        from Ganga.Core import monitoring_component

        thread_handled_states = ['completed', 'failed']
        for job, state, old_state in zip(monitor_jobs, result, ganga_job_status):
            if monitoring_component:
                if monitoring_component.should_stop():
                    break

            job.backend.statusInfo = state[0]
            job.backend.status = state[1]
            job.backend.actualCE = state[2]
            updated_dirac_status = state[3]
            try:
                job.backend.extraInfo = state[4]
            except Exception as err:
                logger.debug("gxception: %s" % str(err))
                pass
            logger.debug('Job status vector  : ' + job.fqid + ' : ' + repr(state))

            # Is this really catching a real problem?
            if job.status != old_state:
                logger.warning('User changed Ganga job status from %s -> %s' % (str(old_state), job.status))
                continue
            ####################

            if updated_dirac_status == job.status:
                continue

            if updated_dirac_status in thread_handled_states:
                if job.status != 'running':
                    DiracBase._getStateTime(job, 'running')
                    if job.status in ['removed', 'killed']:
                        continue
                    if (job.master and job.master.status in ['removed', 'killed']):
                        continue  # user changed it under us
                    job.updateStatus('running')
                    if job.master:
                        job.master.updateMasterJobStatus()

                getQueues()._monitoring_threadpool.add_function(DiracBase.job_finalisation,
                                                           args=(job, updated_dirac_status),
                                                           priority=5, name="Job %s Finalizing" % job.fqid)
                job.been_queued = True

            else:
                DiracBase._getStateTime(job, updated_dirac_status)
                if job.status in ['removed', 'killed']:
                    continue
                if (job.master and job.master.status in ['removed', 'killed']):
                    continue  # user changed it under us
                job.updateStatus(updated_dirac_status)
                if job.master:
                    job.master.updateMasterJobStatus()

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#




'''SuperB setup of Application and RTHandler Ganga methods'''

import ConfigParser
import datetime
import os
import subprocess
import sys

from Ganga.Core import ApplicationConfigurationError
from Ganga.Core import GangaException
from Ganga.Core import FileWorkspace
from Ganga.GPIDev.Adapters.IApplication import IApplication
from Ganga.GPIDev.Adapters.IRuntimeHandler import IRuntimeHandler
from Ganga.GPIDev.Schema import *
from Ganga.GPIDev.Lib.File import File
from Ganga.Lib.Executable import Executable
from Ganga.Lib.LCG import LCG
from Ganga.Lib.Mergers.Merger import TextMerger
from Ganga.Utility.Config import *
import Ganga.Utility.logging

import db
import SBDatasetManager
import SBInputDataset
import SBOutputDataset
import SBSubmission
import objectid
import utils


class SBApp(IApplication):
    '''This class manages the job submission environment and remote 
    worker node user executable interaction.'''
    
    _schema = Schema(Version(2, 0), {
        'executable' : SimpleItem(defvalue='',
                               typelist=['str'],
                               doc='Executable *relative* path after analysis \
                               software unpacking. eg: analysisExe.sh'),
        'software_dir' : SimpleItem(defvalue='', 
                                typelist=['str'], 
                                doc='Local absolute path to analysis \
                                software directory. eg: /home/user/software')
        } )
    _category = 'applications'
    _exportmethods = []
    _name = 'SBApp'
    
    def __init__(self):
        super(SBApp, self).__init__()
    
    def master_configure(self):
        '''This method creates the tar.bz2 archive of user sw directory.
        Such a method is called one time per master job'''
        
        logger.debug('SBApp master_configure called.')
        
        self.now = datetime.datetime.now().strftime("%Y%m%d")
        self.os_arch = os.environ['SBROOT'].split('/')[-1]
        self.user_id = utils.getOwner()
        
        j = self.getJobObject()
        
        # check the target SE status using gridmon DB (updated by nagios monitoring system)
        sql = 'SELECT se_host, nagios_test_service FROM se WHERE name_grid = %s'
        local_SE = db.gridmon(sql, (getConfig('SuperB')['submission_site'], ))
        if local_SE[0]['nagios_test_service'] == 'CRITICAL':
            raise GangaException('Local storage element %s is down.' % local_SE[0]['se_host'])
        #   logger.error('Local storage element %s seems died for gridmon.' % local_SE[0]['se_host'])
        #else:
        #    logger.error('Local storage element %s is back alive for gridmon. !! uncomment exception !!' % local_SE[0]['se_host'])
        
        # create the software directory
        if self.software_dir != '':
            if not os.path.isdir(self.software_dir):
                raise ApplicationConfigurationError(None, 'software_dir must be a directory.')
            
            # make the tar file and update sw_archive parameter
            self.software_dir = os.path.normpath(self.software_dir)
            (head, tail) = os.path.split(self.software_dir)
            self.filename = tail
            self.sw_archive = os.path.join(j.inputdir, tail + '.tar.bz2')
            
            logger.info('Creating archive: %s ...', self.sw_archive)
            logger.info('From: %s', head)
            logger.info('Of: %s', tail)
            
            #savedir = os.getcwd()
            #os.chdir(self.software_dir)
            
            #retcode = subprocess.call("tar -cjf %s * 2>/dev/null" % self.sw_archive, shell=True)
            retcode = subprocess.call("tar -cjf %s -C %s %s 2>/dev/null" % (self.sw_archive, head, tail), shell=True)
            if retcode < 0:
                raise ApplicationConfigurationError(None, 'Error %d while creating archive.' % retcode)
            
            #os.chdir(savedir)
        else:
            raise ApplicationConfigurationError(None, 'software_dir cannot be empty.')
        
        if self.executable == '':
            raise ApplicationConfigurationError(None, 'executable cannot be empty.')
        
        # checking that j.inputdata is a valid object
        if not isinstance(j.inputdata, (SBInputDataset.SBInputPersonalProduction,
                                        SBInputDataset.SBInputProductionAnalysis,
                                        SBInputDataset.SBInputPureAnalysis)):
            msg = 'j.inputdata %s is not allowed' % str(type(j.inputdata))
            raise ApplicationConfigurationError(None, msg)
        
        # checking that j.inputdata (the input dataset) is a valid dataset
        j.inputdata.check()
        
        # checking that j.outputdata (the output dataset) is valid
        if isinstance(j.outputdata, SBOutputDataset.SBOutputDataset):
            j.outputdata.check()
        
        # creating temp dataset
        self.temp_dataset = str(objectid.ObjectId())
        free_string = '%s_%s_%s' % (j.id, j.name, self.filename)
        
        sql = '''INSERT INTO analysis_dataset
            (owner, dataset_id, session, parameters, status)
            VALUES (%s, decode(%s, 'hex'), %s, %s, 'temp');
            
            INSERT INTO analysis_dataset_site
            (dataset_id, site)
            VALUES (decode(%s, 'hex'), %s);'''
        params = (utils.getOwner(), 
            self.temp_dataset, 
            'analysis', 
            {'free_string': free_string},
            self.temp_dataset,
            getConfig('SuperB')['submission_site'])
        db.write(sql, params)
        
        # merger
        j.merger = TextMerger()
        j.merger.files.extend(['severus.log', 'output_files.txt'])
        j.merger.ignorefailed = True
        j.merger.compress = True
        
        j.splitter = SBSubmission.SBSubmission()
        
        return (0, None)
    
    def configure(self, masterappconfig):
        '''It creates the SuperB job wrapper config file and manages application 
        string in Executable() method to be ready to launch the job wrapper.
        It also configures the input and output sandbox. Such a method is called one
         time per master job.'''
        
        logger.debug('SBApp configure called.')
        
        job = self._getParent()
        masterjob = job._getParent()
        
        # creating the configuration file needed by the job wrapper
        config = ConfigParser.RawConfigParser()
        
        config.add_section('REST')
        config.set('REST', 'PROXYVAR', 'X509_USER_PROXY')
        config.set('REST', 'NAME', 'restsbk5')
        config.set('REST', 'HOSTNAME', 'bbr-serv09.cr.cnaf.infn.it')
        config.set('REST', 'PORT', '8443')
        
        config.add_section('EXPORT_VARS')
        config.set('EXPORT_VARS', 'LCG_GFAL_INFOSYS', 'egee-bdii.cnaf.infn.it:2170')
        
        config.add_section('OPTIONS')
        config.set('OPTIONS', 'MODULENAME', 'Analysis')
        # setting PRODSERIES and REQUESTNAME as workaround (this parameters are useless for analysis)
        # TODO: fix asap modifying the job wrapper
        config.set('OPTIONS', 'PRODSERIES', '2011_Full_grid_test')
        config.set('OPTIONS', 'REQUESTNAME', '6b88b34759fab9e2d7e8efe73636dc35')
        config.set('OPTIONS', 'OS_ARCH', self.os_arch)
        
        config.add_section('SOFTWARE')
        config.set('SOFTWARE', 'EXEPATH', self.executable)
        (head, tail) = os.path.split(self.sw_archive)
        config.set('SOFTWARE', 'SWPATH', tail)
        
        config.add_section('INPUT')
        config.set('INPUT', 'INPUTMODE', job.inputdata.input_mode)
        if job.inputdata.input_mode == 'none':
            path = ''
        elif job.inputdata.input_mode == 'list':
            (head, path) = os.path.split(job.inputsandbox[0].name)
        elif job.inputdata.input_mode == 'dir':
            path = job.inputdata.input_path[0]
        else:
            raise ApplicationConfigurationError(None, 'input_mode not recognized.')
        config.set('INPUT', 'INPUTPATH', path)
        
        config.add_section('ANALYSIS')
        config.set('ANALYSIS', 'SUBJOB_ID', job.id)
        config.set('ANALYSIS', 'DATASET', self.temp_dataset)
        if isinstance(job.inputdata, SBInputDataset.SBInputPersonalProduction):
            config.set('ANALYSIS', 'TAG', job.inputdata.soft_version)
            config.set('ANALYSIS', 'SIMULATION', job.inputdata.session)
        
        config.add_section('SITE')
        config.set('SITE', 'USELOCALNAME', 'true')
        
        config.add_section('OUTPUT')
        # <identity_certificato>/<dataset_id>/<date_idjob_usertaballsenzaestensione_jobname>/output/subjobs.id_nomefile.root
        dirTree = ("%s/%%DATASET_ID%%/%s_%s_%s" % (self.user_id, self.now, masterjob.id, self.filename))
        if masterjob.name != '':
            dirTree += ("_%s" % masterjob.name)
        dirTree += '/output'
        config.set('OUTPUT', 'LFN', 'lfn:/grid/superbvo.org/analysis/' + dirTree)
        config.set('OUTPUT', 'OUTPUTPATH', dirTree)
        config.set('OUTPUT', 'LOG', '')
        
        config.add_section('OUTPUT_DATASET')
        if isinstance(job.outputdata, SBOutputDataset.SBOutputDataset):
            for key, value in job.outputdata.pairs.items():
                config.set('OUTPUT_DATASET', key, value)
        
        config.add_section('TARGETSITE')
        config.set('TARGETSITE', 'SITENAME', getConfig('SuperB')['submission_site'])
        config.set('TARGETSITE', 'USELOCALNAME', 'false')
        
        fileConf = os.path.join(job.inputdir, "file_%d.conf" % job.id)
        f = open(fileConf, "w")
        config.write(f)
        f.close()
        
        severus_dir = getConfig('SuperB')['severus_dir']
        severus = os.path.join(severus_dir, 'severus.tgz')
        severus_wrapper = os.path.join(severus_dir, 'severus_wrapper.sh')
        
        # updating the information of application, inputsandbox and outputsandbox  objects 
        # (the job wrapper, the configuration file, the user software, run number and 
        # verbosity flag as command line parameters, the output log and a list of output files)
        job.application = Executable()
        job.application.exe = File(severus_wrapper)
        # severus verbose (the string 'test' as third parameter)
        job.application.args = [File(fileConf), 10000000, 'test']
        # severus quiet (missing third parameter)
        #job.application.args = [File(fileConf), 10000000]
        job.inputsandbox.append(File(severus))
        job.inputsandbox.append(File(self.sw_archive))
        job.outputsandbox.append('severus.log')
        job.outputsandbox.append('output_files.txt')
        
        return (None, None)




def convertIntToStringArgs(args):
    result = []
    for arg in args:
        if isinstance(arg,int):
            result.append(str(arg))
        else:
            result.append(arg)
    return result

class RTHandler(IRuntimeHandler):
    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        logger.debug('RTHandler prepare called.')
        
        from Ganga.GPIDev.Adapters.StandardJobConfig import StandardJobConfig
        return StandardJobConfig(app.exe,
                                 app._getParent().inputsandbox,
                                 convertIntToStringArgs(app.args),
                                 app._getParent().outputsandbox,
                                 app.env)


class LCGRTHandler(IRuntimeHandler):
    
    def master_prepare(self, app, appmasterconfig):
        #Get the string jdl requirement per site and set it up for ganga job
        
        logger.debug('LCGRTHandler master_prepare called.')
        j = app._getParent()
        
        if len(j.inputdata.run_site) == 0:
            j.inputdata.getDefaultRunSite()
        
        # check CEs status  of the chosen site using gridmon DB (updated by nagios monitoring system)
        sql = '''SELECT name_grid, 
                status_other, 
                banned 
            FROM site 
            WHERE false'''
        for site in j.inputdata.run_site:
            sql += ' OR name_grid = \'%s\' ' % site
        sql += 'ORDER BY name_grid'
        
        sites = db.gridmon(sql)
        
        for site in sites:
            if site['banned'] == True:
                logger.error('%s is not available due to technical problems.' % site['name_grid'])
                sites.remove(site)
            
            if site['status_other'] != 'OK':
                logger.error('%s is not available because of administrative policies.' % site['name_grid'])
                sites.remove(site)
        
        if len(sites) == 0:
            raise ApplicationConfigurationError(None, 'No sites available, try later again.')
        
        sql = '''SELECT ce_host
            FROM ce
            WHERE (nagios_test_service = 'OK' OR nagios_test_service = 'WARNING') AND (false'''
        for site in sites:
            sql += ' OR name_grid = \'%s\'' % site['name_grid']
        sql += ' ) ORDER BY name_grid, ce_host'
        
        ces = db.gridmon(sql)
        
        requirements = 'False'
        
        for ce in ces:
            requirements += ' || '
            requirements += '(other.GlueCEInfoHostName == "%s")' % ce['ce_host']
        
        # converting from unicode string to normal python string (why is it necessary?)
        requirements = str(requirements)
        
        # cleaning useless flags
        requirements = requirements.replace('False || ', '')
        
        # debugging print
        logger.debug('Requirements: %s' % requirements)
        
        # if all sites are down and/or requirements are not defined, stop submission
        if requirements == 'False':
            raise ApplicationConfigurationError(None, 'Requirements are empty.')
        
        j.backend.requirements.other += [requirements]
        
        # setting WALLTIME (if the user hasn't setted it before)
        if j.backend.requirements.walltime == 0:
            j.backend.requirements.walltime = 24 * 60 # dobbiamo fornire un valore in minuti
        
        #from Ganga.Lib.LCG import LCGJobConfig
        #return LCGJobConfig(app.exe,app._getParent().inputsandbox,convertIntToStringArgs(app.args),app._getParent().outputsandbox,app.env)
    
    def prepare(self, app, appconfig, appmasterconfig, jobmasterconfig):
        logger.debug('LCGRTHandler prepare called.')
        
        job = app._getParent()
        masterjob = job._getParent()
        
        job.backend.requirements.other = masterjob.backend.requirements.other
        job.backend.requirements.walltime = masterjob.backend.requirements.walltime
        
        from Ganga.Lib.LCG import LCGJobConfig
        return LCGJobConfig(app.exe, 
                            app._getParent().inputsandbox, 
                            convertIntToStringArgs(app.args), 
                            app._getParent().outputsandbox, 
                            app.env)


class gLiteRTHandler(IRuntimeHandler):
    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        logger.debug('gLiteRTHandler prepare called.')
        
        from Ganga.Lib.gLite import gLiteJobConfig
        return gLiteJobConfig(app.exe,
                              app._getParent().inputsandbox,
                              convertIntToStringArgs(app.args),
                              app._getParent().outputsandbox,
                              app.env)


from Ganga.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers


allHandlers.add('SBApp','LSF', RTHandler)
allHandlers.add('SBApp','Local', RTHandler)
allHandlers.add('SBApp','PBS', RTHandler)
allHandlers.add('SBApp','SGE', RTHandler)
allHandlers.add('SBApp','Condor', RTHandler)
allHandlers.add('SBApp','LCG', LCGRTHandler)
allHandlers.add('SBApp','gLite', gLiteRTHandler)
allHandlers.add('SBApp','TestSubmitter', RTHandler)
allHandlers.add('SBApp','Interactive', RTHandler)
allHandlers.add('SBApp','Batch', RTHandler)
allHandlers.add('SBApp','Cronus', RTHandler)
allHandlers.add('SBApp','Remote', LCGRTHandler)
allHandlers.add('SBApp','CREAM', LCGRTHandler)

logger = Ganga.Utility.logging.getLogger()



from Ganga.GPIDev.Base import GangaObject
from Ganga.GPIDev.Schema import *

import Ganga.Utility.logging
logger = Ganga.Utility.logging.getLogger()

from Ganga.Utility.Config import makeConfig,getConfig

from Ganga.GPIDev.Lib.Job import JobError

monconf = getConfig('PollThread')
monconf.addOption('TestSubmitter',1,'poll rate for test submitter')


## # test configuration properties
## test_config = makeConfig('TestConfig','testing stuff')
## test_config.addOption('None_OPT', None, '')
## test_config.addOption('Int_OPT', 1, '')
## test_config.addOption('List_OPT', [1,2,3], '')
## test_config.addOption('String_OPT' ,'dupa', '')
## # there is also an Undefine_OPT which will be used in the test case

import time

from Ganga.GPIDev.Adapters.IBackend import IBackend

class TestSubmitter(IBackend):
    _schema = Schema(Version(1,0), {'time' : SimpleItem(defvalue=5,changable_at_resubmit=1),
                                    'start_time' : SimpleItem(defvalue=0,protected=1,comparable=0,copyable=0),
                                    'update_delay' : SimpleItem(defvalue=0,doc="The time it takes to updateMonitoringInformation"),
                                    'fail' : SimpleItem(defvalue='',doc='Define the artificial runtime failures: "submit", "kill","monitor","resubmit"', changeable_at_resubmit=1 )
                                    
                                    })
    _category = 'backends'
    _name = 'TestSubmitter'

    def __init__(self):
        super(TestSubmitter,self).__init__()

    def tryfail(self,what):
        if self.fail == what:

            logger.info('triggered failure during %s',what)
            x = 'triggered failure during %s'%what
            raise JobError(x)
        
    def submit(self,jobconfig,masterjobconfig):
        jobid = self.getJobObject().getFQID('.')
        logger.info('testing submission of job %s',jobid)
        logger.info('this job will be finished in approx. %d seconds',self.time)
        self.start_time = time.time()
        self.tryfail('submit')
        return 1

    def resubmit(self,backend=None):
        if backend:
            self.time = backend.time

        jobid = self.getJobObject().getFQID('.')
        logger.info('testing resubmission of job %s',jobid)
        logger.info('this job will be finished in approx. %d seconds',self.time)
        self.start_time = time.time()
        self.tryfail('resubmit')
        return 1        

    def kill(self):
        jobid = self.getJobObject().getFQID('.')
        r = self.remaining()
        logger.info('testing killing of job %s, remaining seconds to complete %d',jobid,r)
        self.tryfail('kill')
        self.time = 0
        return r>0
        
    def remaining(self):
        return self.time-(time.time()-self.start_time)        
    
    @staticmethod
    def updateMonitoringInformation(jobs):
        logger.info('monitoring jobs=%s',jobs)
        for j in jobs:
            logger.info('monitoring id=%d',j.id)
            if j.backend.update_delay:
                logger.info('job %d: updateMonitoringInformation sleeping for %d s'%(j.id,j.backend.update_delay))
                time.sleep(j.backend.update_delay)
            j.backend.tryfail('monitor')
            if j.backend.remaining() <= 0:
                logger.info('job %d completed',j.id)
                j.updateStatus('completed')
            else:
                j.updateStatus('running')
                logger.info('job %d seconds remaining %d',j.id,j.backend.remaining())




from __future__ import absolute_import
##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: JobRegistry.py,v 1.1.2.1 2009-07-24 13:39:39 ebke Exp $
##########################################################################

#from Ganga.Utility.external.ordereddict import oDict
from Ganga.Utility.external.OrderedDict import OrderedDict as oDict

from Ganga.Core.exceptions import GangaException
from Ganga.Core.GangaRepository.Registry import Registry, RegistryKeyError, RegistryAccessError, RegistryFlusher

from Ganga.GPIDev.Base.Proxy import stripProxy, isType, addProxy

import Ganga.Utility.logging

from Ganga.GPIDev.Lib.Job.Job import Job

from .RegistrySlice import RegistrySlice

from .RegistrySliceProxy import RegistrySliceProxy, _wrap, _unwrap

# display default values for job list
from .RegistrySlice import config

logger = Ganga.Utility.logging.getLogger()


class JobRegistry(Registry):

    def __init__(self, name, doc, dirty_flush_counter=10, update_index_time=30, dirty_max_timeout=60, dirty_min_timeout=30):
        super(JobRegistry, self).__init__(name, doc, dirty_flush_counter, update_index_time, dirty_max_timeout, dirty_min_timeout)
        self.stored_slice = JobRegistrySlice(self.name)
        self.stored_slice.objects = self
        self.stored_proxy = JobRegistrySliceProxy(self.stored_slice)

    def getSlice(self):
        return self.stored_slice

    def getProxy(self):
        return self.stored_proxy

    def getIndexCache(self, obj):

        cached_values = ['status', 'id', 'name']
        cache = {}
        for cv in cached_values:
            #print("cv: %s" % str(cv))
            cache[cv] = getattr(obj, cv)
            #logger.info("Setting: %s = %s" % (str(cv), str(cache[cv])))
        this_slice = JobRegistrySlice("jobs")
        for dpv in this_slice._display_columns:
            #logger.debug("Storing: %s" % str(dpv))
            try:
                value = this_slice._get_display_value(obj, dpv)
                cache["display:" + dpv] = value
            except Exception as err:
                value = None
        del this_slice

        # store subjob status
        if hasattr(obj, "subjobs"):
            cache["subjobs:status"] = []
            if hasattr(obj.subjobs, "getAllCachedData"):
                for sj in obj.subjobs.getAllCachedData():
                    cache["subjobs:status"].append(sj['status'])
            else:
                for sj in obj.subjobs:
                    cache["subjobs:status"].append(sj.status)

        #print("Cache: %s" % str(cache))
        return cache

    def startup(self):
        self._needs_metadata = True
        super(JobRegistry, self).startup()
        if len(self.metadata.ids()) == 0:
            from Ganga.GPIDev.Lib.JobTree import JobTree
            jt = JobTree()
            stripProxy(jt)._setRegistry(self.metadata)
            self.metadata._add(jt)
        self.jobtree = self.metadata[self.metadata.ids()[-1]]
        self.flush_thread = RegistryFlusher(self)
        self.flush_thread.start()

    def shutdown(self):
        self.flush_thread.join()
        super(JobRegistry, self).shutdown()

    def getJobTree(self):
        return self.jobtree

    def _remove(self, obj, auto_removed=0):
        super(JobRegistry, self)._remove(obj, auto_removed)
        try:
            self.jobtree.cleanlinks()
        except Exception as err:
            logger.debug("Exception in _remove: %s" % str(err))
            pass

class JobRegistrySlice(RegistrySlice):

    def __init__(self, name):
        super(JobRegistrySlice, self).__init__(name, display_prefix="jobs")
        from Ganga.Utility.ColourText import Foreground, Background, Effects
        fg = Foreground()
        fx = Effects()
        bg = Background()
        try:
            status_colours = config['jobs_status_colours']
            self.status_colours = dict([(k, eval(v)) for k, v in status_colours.iteritems()])
        except Exception as x:
            logger.warning('configuration problem with colour specification: "%s"', str(x))
            status_colours = config.options['jobs_status_colours'].default_value
            self.status_colours = dict([(k, eval(v)) for k, v in status_colours.iteritems()])
        self.fx = fx
        self._proxyClass = JobRegistrySliceProxy

    def _getColour(self, obj):
        if isType(obj, Job):
            from Ganga.GPIDev.Lib.Job.Job import lazyLoadJobStatus
            status_attr = lazyLoadJobStatus(stripProxy(obj))
        elif isType(obj, str):
            status_attr = obj
        elif isType(obj, dict):
            if 'display:status' in obj.keys():
                status_attr = obj['display:status']
            elif 'status' in obj.keys():
                status_attr = obj['status']
            else:
                status_attr = None
        else:
            status_attr = obj
        try:
            returnable = self.status_colours.get(status_attr, self.fx.normal)
        except Exception:
            returnable = self.status_colours.get(self.fx.normal)
        return returnable

    def __call__(self, id):
        """ Retrieve a job by id.
        """
        this_id = id
        t = type(this_id)
        if t is int:
            try:
                return _wrap(self.objects[this_id])
            except KeyError:
                if self.name == 'templates':
                    raise RegistryKeyError('Template %d not found' % this_id)
                else:
                    raise RegistryKeyError('Job %d not found' % this_id)
        elif t is tuple:
            ids = this_id
        elif t is str:
            if this_id.isdigit():
                try:
                    return _wrap(self.objects[int(this_id)])
                except KeyError:
                    if self.name == 'templates':
                        raise RegistryKeyError('Template %d not found' % this_id)
                    else:
                        raise RegistryKeyError('Job %d not found' % this_id)
            elif this_id.count('.') == 1 and id.split('.')[0].isdigit() and this_id.split('.')[1].isdigit():
                ids = this_id.split(".")
            else:
                import fnmatch
                jlist = [j for j in self.objects if fnmatch.fnmatch(j.name, this_id)]
                if len(jlist) == 1:
                    return _wrap(jlist[0])
                return jobSlice(jlist)
        else:
            raise RegistryAccessError('Expected a job id: int, (int,int), or "int.int"')

        if not len(ids) in [1, 2]:
            raise RegistryAccessError('Too many ids in the access tuple, 2-tuple (job,subjob) only supported')

        try:
            ids = [int(this_id) for this_id in ids]
        except TypeError:
            raise RegistryAccessError('Expeted a job id: int, (int,int), or "int.int"')
        except ValueError:
            raise RegistryAccessError('Expected a job id: int, (int,int), or "int.int"')

        try:
            j = self.objects[ids[0]]
        except KeyError:
            if self.name == 'templates':
                raise RegistryKeyError('Template %d not found' % ids[0])
            else:
                raise RegistryKeyError('Job %d not found' % ids[0])

        if len(ids) > 1:
            try:
                return _wrap(j.subjobs[ids[1]])
            except IndexError:
                raise RegistryKeyError('Subjob %s not found' % ('.'.join([str(_id) for _id in ids])))
        else:
            return _wrap(j)

    def submit(self, keep_going):
        self.do_collective_operation(keep_going, 'submit')

    def kill(self, keep_going):
        self.do_collective_operation(keep_going, 'kill')

    def resubmit(self, keep_going):
        self.do_collective_operation(keep_going, 'resubmit')

    def force_status(self, status, keep_going, force):
        self.do_collective_operation(
            keep_going, 'force_status', status, force=force)

    def remove(self, keep_going, force):
        self.do_collective_operation(keep_going, 'remove', force=force)


class JobRegistrySliceProxy(RegistrySliceProxy):

    """This object is an access list of jobs defined in Ganga. 

    'jobs' represents all existing jobs, so a subset of jobs may 
    be created by slicing: 
    jobs[-10:]      (to get the last ten jobs)

    or selecting:
    jobs.select(status='new')      (to select all new jobs)
    jobs.select(10,20)      (to select those with IDs between 10 and 20)

    A new access list is created as a result of a slice/select operation. 
    The new access list may be further restricted.

    This object allows one to perform collective operations such as
    killing or submiting all jobs in the current range. Setting the optional
    parameter keep_going=True (the default) results in the operation continuing 
    to process all jobs, irrespective of any errors encountered. If keep_going=False 
    then the operation will stop with an Exception at the first error encountered.


    """

    def submit(self, keep_going=True):
        """ Submit all jobs."""
        stripProxy(self).submit(keep_going=keep_going)

    def resubmit(self, keep_going=True):
        """ Resubmit all jobs."""
        return stripProxy(self).resubmit(keep_going=keep_going)

    def kill(self, keep_going=True):
        """ Kill all jobs."""
        return stripProxy(self).kill(keep_going=keep_going)

    def remove(self, keep_going=True, force=False):
        """ Remove all jobs."""
        return stripProxy(self).remove(keep_going=keep_going, force=force)

    def fail(self, keep_going=True, force=False):
        """ Fail all jobs."""
        return stripProxy(self).fail(keep_going=keep_going, force=force)

    def force_status(self, status, keep_going=True, force=False):
        """ Force status of all jobs to 'completed' or 'failed'."""
        return stripProxy(self).force_status(status, keep_going=keep_going, force=force)

    def copy(self, keep_going=True):
        """ Copy all jobs. """
        return JobRegistrySliceProxy(stripProxy(self).copy(keep_going=keep_going))

    def __call__(self, x):
        """ Access individual job. Examples:
        jobs(10) : get job with id 10 or raise exception if it does not exist.
        jobs((10,2)) : get subjobs number 2 of job 10 if exist or raise exception.
        jobs('10.2')) : same as above
        """
        return stripProxy(self).__call__(x)

    def __getslice__(self, i1, i2):
        """ Get a slice. Examples:
        jobs[2:] : get first two jobs,
        jobs[-10:] : get last 10 jobs.
        """
        return _wrap(stripProxy(self).__getslice__(i1, i2))

    def __getitem__(self, x):
        """ Get a job by positional index. Examples:
        jobs[-1] : get last job,
        jobs[0] : get first job,
        jobs[1] : get second job.
        """
        return _wrap(stripProxy(self).__getitem__(_unwrap(x)))

def jobSlice(joblist):
    """create a 'JobSlice' from a list of jobs
    example: jobSlice([j for j in jobs if j.name.startswith("T1:")])"""
    this_slice = JobRegistrySlice("manual slice")
    this_slice.objects = oDict([(j.fqid, _unwrap(j)) for j in joblist])
    return _wrap(this_slice)

# , "Create a job slice from a job list") exported to the Runtime bootstrap




from Ganga.GPIDev.Lib.Tasks.common import getJobByID
from Ganga.GPIDev.Lib.Tasks.ITransform import ITransform
from Ganga.GPIDev.Lib.Job.Job import JobError
from Ganga.GPIDev.Lib.Registry.JobRegistry import JobRegistrySlice, JobRegistrySliceProxy
from Ganga.Core.exceptions import ApplicationConfigurationError
from Ganga.GPIDev.Lib.Tasks.ITransform import ITransform
from Ganga.GPIDev.Lib.Tasks.TaskLocalCopy import TaskLocalCopy
from Ganga.GPIDev.Lib.File.MassStorageFile import MassStorageFile
from ND280Unit import ND280Unit
from modules.ND280Dataset.ND280Dataset import ND280LocalDataset, ND280DCacheDataset
from modules.ND280Splitter.ND280Splitter import splitNbInputFile
import os

class ND280Transform(ITransform):
   _schema = Schema(Version(1,0), dict(ITransform._schema.datadict.items() + {
        'nbinputfiles' : SimpleItem(defvalue=1,doc='The max number of files assigned to each unit. Use 0 to put all the available files in each given inputdata in one unit (i.e. N inputdata => N units).'),
        'inputdatasubsets' : SimpleItem(defvalue=[], hidden=1,doc='List of subsets of files. The number of files in each subset is equal to nbinputfiles'),
    }.items()))

   _category = 'transforms'
   _name = 'ND280Transform'
   _exportmethods = ITransform._exportmethods + [ ]


   def __init__(self):
      super(ND280Transform,self).__init__()
      self.inputdatasubsets = []

   def createUnits(self):
      """Create new units if required given the inputdata"""
      
      # call parent for chaining
      super(ND280Transform,self).createUnits()
      
      # loop over input data and see if we need to create any more units
      for idx,inds in enumerate(self.inputdata):
         
        # currently only checking for Local Datasets and DCache
        if inds._name == "ND280LocalDataset" or inds._name == "ND280DCacheDataset":
           # First pass, create all the subsets
           if len(self.inputdatasubsets) == idx:
             subsets = splitNbInputFile(inds.names, self.nbinputfiles)
             self.inputdatasubsets.append(subsets)
           elif len(self.inputdatasubsets) < idx:
             raise Exception('ND280Transform: The inputdata and inputdatasubsets are out of sync. This should not happen.')

           ok = True

           for subset in self.inputdatasubsets[idx]:
              # check if this data is being run over by checking all the names listed
              ok = False
              for unit in self.units:
                 if (set(unit.inputdata.names) == set(subset)):
                    ok = True
                    break

              if not ok:
                 # new unit required for this subset
                 unit = ND280Unit()
                 unit.name = "Unit %d" % len(self.units)
                 self.addUnitToTRF( unit )
                 if inds._name == "ND280LocalDataset":
                    unit.inputdata = ND280LocalDataset()
                 elif inds._name == "ND280DCacheDataset":
                    unit.inputdata = ND280DCacheDataset()
                 unit.inputdata.names = subset

      # For special cases where there is no inputdata given,
      # just create one unit.
      if len(self.inputdata) == 0:
         unit = ND280Unit()
         unit.name = "Unit %d" % len(self.units)
         self.addUnitToTRF( unit )


   def addUnit(self, filelist):
      """Create a new unit based on this file list"""
      unit = ND280Unit()
      unit.inputdata = ND280LocalDataset()
      unit.inputdata.names = filelist
      self.addUnitToTRF( unit )

   def createChainUnit( self, parent_units, use_copy_output = True ):
      """Create a chained unit using the output data from the given units"""

      # check all parent units for copy_output
      copy_output_ok = True
      for parent in parent_units:
         if not parent.copy_output:
            copy_output_ok = False

      # all parent units must be completed so the outputfiles are filled correctly
      for parent in parent_units:
         if parent.status != "completed":
           return None

      if len(parent_units) == 0:
         return None

      if not use_copy_output or not copy_output_ok:
         unit = ND280Unit()
         unit.inputdata = ND280LocalDataset()
         for parent in parent_units:
            # loop over the output files and add them to the ND280LocalDataset - THIS MIGHT NEED SOME WORK!
            job = getJobByID(parent.active_job_ids[0])
            for f in job.outputfiles:
               # should check for different file types and add them as appropriate to the dataset
               # self.inputdata (== TaskChainInput).include/exclude_file_mask could help with this
               # This will be A LOT easier with Ganga 6.1 as you can easily map outputfiles -> inputfiles!
               # TODO: implement use of include/exclude_file_mask
               #       
               try:
                 outputfilenameformat = f.outputfilenameformat
               except:
                 inputdir = job.outputdir
               else:
                 #### WARNING: The following will work only if the MassStorageFile puts the files in local directories !
                 inputdir = '/'.join( [getConfig('Output')['MassStorageFile']['uploadOptions']['path'], f.outputfilenameformat.replace('{fname}','')])
               unit.inputdata.get_dataset( inputdir, f.namePattern )
      else:

         unit = ND280Unit()
         unit.inputdata = ND280LocalDataset()

         for parent in parent_units:
            # unit needs to have completed and downloaded before we can get file list
            if parent.status != "completed":
               return None

            # we should be OK so copy all output to the dataset
            for f in parent.copy_output.files:
               unit.inputdata.names.append( os.path.join( parent.copy_output.local_location, f ) )
         
      return unit
   



#!/usr/bin/env python

import os,sys
opts = '###OPTS###'
project_opts = '###PROJECT_OPTS###'
app = '###APP_NAME###'
app_upper = app.upper()
version = '###APP_VERSION###'
package = '###APP_PACKAGE###'
platform = '###PLATFORM###'


# check that options file exists
if not os.path.exists(opts):
    opts = 'notavailable'
    os.environ['JOBOPTPATH'] = opts
else:
    os.environ['JOBOPTPATH'] = os.path.join(os.environ[app + '_release_area'],
            app_upper,
            app_upper,
            version,
            package,
            app,
            version,
            'options',
            'job.opts')
    sys.stdout.write('Using the master optionsfile: %s' % opts)
    sys.stdout.flush()

# # # # # Code to Setup Environment # # # # #
###CONSTRUCT_ENVIRON###
# # # # Code to Setup Environment # # # # #

# add lib subdir in case user supplied shared libs where copied to pwd/lib
os.environ['LD_LIBRARY_PATH'] = '.:%s/lib:%s' % (os.getcwd(), os.environ['LD_LIBRARY_PATH'])

#run
sys.stdout.flush()
os.environ['PYTHONPATH'] = '%s/InstallArea/python:%s' % \
        (os.getcwd(), os.environ['PYTHONPATH'])
os.environ['PYTHONPATH'] = '%s/InstallArea/%s/python:%s' % \
        (os.getcwd(), platform,os.environ['PYTHONPATH'])

cmdline = '''###CMDLINE###'''

# run command
os.system(cmdline)

###XMLSUMMARYPARSING###




##########################################################################
# Ganga - a computational task management tool for easy access to Grid resources
# http://cern.ch/ganga
#
# $Id: Coordinator.py,v 1.1.4.2 2009-07-14 14:44:17 ebke Exp $
#
# Copyright (C) 2003-2007 The Ganga Project
#
# This file is part of Ganga.
#
# Ganga is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Ganga is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
##########################################################################

"""
 Internal services coordinator :
  takes care of conditional enabling/disabling of internal services (job monitoring loop, job registry, job
  repository/workspace) when credentials become invalid preventing normal functioning of these services.
  E.g: invalid grid proxy triggers the monitor-loop stop
"""
from Ganga.Utility.Config import getConfig
from Ganga.Utility.logging import getLogger
from Ganga.GPIDev.Base.Proxy import getName

log = getLogger()

# the overall state of Ganga internal services
servicesEnabled = True


def isCredentialRequired(credObj):
    """
    The logic to decide if a given invalid credential
    should trigger the deactivation of Ganga internal services.  
    """

    from Ganga.Runtime import Workspace_runtime
    from Ganga.Runtime import Repository_runtime

    if getName(credObj) == 'AfsToken':
        return Workspace_runtime.requiresAfsToken() or Repository_runtime.requiresAfsToken()

    if getName(credObj) == 'GridProxy':
        from Ganga.Core.GangaRepository import getRegistryProxy
        from Ganga.Runtime.GPIFunctions import typename
        from Ganga.GPIDev.Base.Proxy import stripProxy
        from Ganga.GPIDev.Lib.Job.Job import lazyLoadJobBackend, lazyLoadJobStatus
        for j in getRegistryProxy('jobs'):
            ji = stripProxy(j)
            this_status = lazyLoadJobStatus(ji)
            if this_status in ['submitted', 'running', 'completing']:
                this_backend = lazyLoadJobBackend(ji)
                if getName(this_backend) == 'LCG':
                    return True
        return False

    log.warning("Unknown credential object : %s" % credObj)


def notifyInvalidCredential(credObj):
    """
    The Core is notified when one of the monitored credentials is invalid
    @see ICredential.create()
    """

    # ignore this notification if the internal services are already stopped
    if not servicesEnabled:
        log.debug(
            "One of the monitored credential [%s] is invalid BUT the internal services are already disabled." % getName(credObj))
        return

    if isCredentialRequired(credObj):
        log.debug("One of the required credential for the internal services is invalid: [%s]."
                  "Disabling internal services ..." % getName(credObj))
        _tl = credObj.timeleft()
        if _tl == "-1":
            log.error('%s has been destroyed! Could not shutdown internal services.' % getName(credObj))
            return
        disableInternalServices()
        log.warning('%s is about to expire! '
                    'To protect against possible write errors all internal services has been disabled.'
                    'If you believe the problem has been solved type "reactivate()" to re-enable '
                    'interactions within this session.' % getName(credObj))
    else:
        log.debug("One of the monitored credential [%s] is invalid BUT it is not required by the internal services" % getName(credObj))


def _diskSpaceChecker():
    """
    the callback function used internally by Monitoring Component
    Reads and calls the checking function provided in the configuration. 
    If this checking function returns "False" the internal services are disabled making Ganga read-only:
    e.g:
    [PollThread]
    DiskSpaceChecker =  
        import commands
        diskusage = commands.getoutput('df -l -P %s/workspace' % config['Configuration']['gangadir'])
        used  = diskusage.splitlines()[1].split()[4] # get disk usage (in %)
        return int(used[:-1])<70
    """
    log.debug("Checking disk space")
    try:
        config = getConfig('PollThread')

        if config['DiskSpaceChecker']:
            _checker = lambda: True
            try:
                # create the checker
                from Ganga.Runtime import _prog
                import new
                ns = {}
                code = "def check():"
                for line in config['DiskSpaceChecker'].splitlines():
                    code += "\t%s\n" % line
                exec code in ns
                _checker = new.function(
                    ns["check"].__code__, _prog.local_ns, 'check')
            except Exception as e:
                log.warning(
                    'Syntax errors in disk space checking code: %s. See [PollThread]DiskSpaceChecker' % e)
                return False

            # call the checker
            if _checker() is False:
                disableInternalServices()
                log.warning('You are running out of disk space! '
                            'To protect against possible write errors all internal services has been disabled.'
                            'If you believe the problem has been solved type "reactivate()" to re-enable '
                            'interactions within this session.')
    except Exception as msg:
        log.warning(
            'Exception in free disk space checking code: %s. See [PollThread]DiskSpaceChecker' % msg)
        return False
    return True


def disableMonitoringService():

    # disable the mon loop
    log.debug("Shutting down the main monitoring loop")
    from Ganga.Core.MonitoringComponent.Local_GangaMC_Service import _purge_actions_queue, stop_and_free_thread_pool
    _purge_actions_queue()
    stop_and_free_thread_pool()
    log.debug("Disabling the central Monitoring")
    from Ganga.Core import monitoring_component
    monitoring_component.disableMonitoring()


def disableInternalServices():
    """
    Deactivates all the internal services :
          * monitoring loop
          * registry/repository and workspace (or GPI entierly)
    Currently this method is called whenever:
          * one of the managed credentials (AFS token or Grid Proxy) is detected as beeing *invalid* by the monitoring component
          * the user is running out of space
    """

    log.info("Ganga is now attempting to shut down all running processes accessing the repository in a clean manner")
    log.info(" ... Please be patient! ")


    ## MOVED TO THE END OF THE SHUTDOWN SO THAT WE NEVER ACCESS A REPO BEFORE WE ARE FINISHED!
    # flush the registries
    #log.debug("Coordinator Shutting Down Repository_runtime")
    #from Ganga.Runtime import Repository_runtime
    #Repository_runtime.shutdown()

    global servicesEnabled

    if not servicesEnabled:
        log.error("Cannot disable services, they've already been disabled")
        from Ganga.Core.exceptions import GangaException
        raise GangaException("Cannot disable services")

    log.debug("Disabling the internal services")

    # disable the mon loop
    disableMonitoringService()

    # For debugging what services are still alive after being requested to stop before we close the repository
    #from Ganga.Core.MonitoringComponent.Local_GangaMC_Service import getStackTrace
    # getStackTrace()
    # log.info(queues_threadpoolMonitor._display(0))

    log.debug("Ganga is now about to shutdown the repository, any errors after this are likely due to badly behaved services")

    log.info("Ganga is shutting down the repository, to regain access, type 'reactivate()' at your prompt")

    # flush the registries
    #log.debug( "Coordinator Shutting Down Repository_runtime" )
    #from Ganga.Runtime import Repository_runtime
    # Repository_runtime.shutdown()

    # this will disable any interactions with the registries (implicitly with
    # the GPI)
    servicesEnabled = False


def enableMonitoringService():
    from Ganga.Core import monitoring_component
    monitoring_component.alive = True
    monitoring_component.enableMonitoring()
    from Ganga.Core.MonitoringComponent.Local_GangaMC_Service import _makeThreadPool, ThreadPool
    if not ThreadPool or len(ThreadPool) == 0:
        _makeThreadPool()
    global servicesEnabled
    servicesEnabled = True

def enableInternalServices():
    """
    activates the internal services previously disabled due to expired credentials
    """
    global servicesEnabled

    if servicesEnabled:
        log.error("Cannot (re)enable services, they're already running")
        from Ganga.Core.exceptions import GangaException
        raise GangaException("Cannot (re)enable services")

    # startup the registries
    from Ganga.Runtime import Repository_runtime
    Repository_runtime.bootstrap()

    # make sure all required credentials are valid
    missing_cred = getMissingCredentials()
    if missing_cred:
        log.error("The following credentials are still required: %s."
                  "Make sure you renew them before reactivating this session" % ','.join(missing_cred))
        return

    log.debug("Enabling the internal services")
    # re-enable the monitoring loop as it's been explicityly requested here
    enableMonitoringService()

    servicesEnabled = True
    log.info('Internal services reactivated successfuly')


def checkInternalServices(errMsg='Internal services disabled. Job registry is read-only.'):
    """
    Check the state of internal services and return a ReadOnlyObjectError exception
    in case the state is disabled.    
    """

    global servicesEnabled
    from Ganga.GPIDev.Base import ReadOnlyObjectError

    if not servicesEnabled:
        raise ReadOnlyObjectError(errMsg)


def getMissingCredentials():
    """
    get a list of missing credentials
    i.e:  invalid credentials that are needed by the internal services to run
    """
    from Ganga.GPIDev.Credentials import _allCredentials as availableCreds
    return [name for name in availableCreds
            if not availableCreds[name].isValid() and
            isCredentialRequired(availableCreds[name])]


def bootstrap():

    #global servicesEnabled
    #servicesEnabled = True

    # export to GPI moved to Runtime bootstrap

    servicesEnabled = True

#
#$Log: not supported by cvs2svn $
# Revision 1.1.4.1  2009/07/08 11:18:21  ebke
# Initial commit of all - mostly small - modifications due to the new GangaRepository.
# No interface visible to the user is changed
#
# Revision 1.1  2008/07/17 16:40:50  moscicki
# migration of 5.0.2 to HEAD
#
# the doc and release/tools have been taken from HEAD
#
# Revision 1.3.6.4  2008/03/11 15:22:42  moscicki
# merge from Ganga-5-0-restructure-config-branch
#
# Revision 1.3.6.3.2.1  2008/03/07 13:36:07  moscicki
# removal of [DefaultJobRepository] and [FileWorkspace]
# new options in [Configuration] user, gangadir, repositorytype, workspacetype
#
# Revision 1.3.6.3  2008/02/12 09:25:52  amuraru
# fixed repositories shutdown
#
# Revision 1.3.6.2  2008/02/05 12:33:23  amuraru
# fixed DiskSpaceChecker alignment
#
# Revision 1.3.6.1  2007/12/10 19:24:42  amuraru
# merged changes from Ganga 4.4.4
#
# Revision 1.7  2007/12/05 12:42:54  amuraru
# Ganga/Core/InternalServices/Coordinator.py
#
# Revision 1.6  2007/12/04 12:59:03  amuraru
#*** empty log message ***
#
# Revision 1.5  2007/11/26 14:04:30  amuraru
# allow indentation using \t tab char in DiskSpaceChecker
#
# Revision 1.4  2007/10/29 14:06:00  amuraru
# added free disk space checker to monitoring loop
#
# Revision 1.3  2007/07/27 18:02:34  amuraru
# updated to comply with the latest requirement for GPI free functions docstrings
#
# Revision 1.2  2007/07/27 14:31:56  moscicki
# credential and clean shutdown updates from Adrian (from Ganga-4-4-0-dev-branch)
#
# Revision 1.1.2.1  2007/07/27 13:04:00  amuraru
#*** empty log message ***
#
#



##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: IRuntimeHandler.py,v 1.1 2008-07-17 16:40:52 moscicki Exp $
##########################################################################

import Ganga.Utility.logging
logger = Ganga.Utility.logging.getLogger()


class IRuntimeHandler(object):

    """ The RuntimeHandler is a connector between the application and the backend.

    Application configure methods produce appconfig objects. Backend submit method
    consumes the jobconfig object. RuntimeHandler translates the appconfig objects
    into the jobconfig objects. The translation is a part of the job submission.
    It is implemented by the prepare methods.

    """

    def master_prepare(self, app, appmasterconfig):
        """ Prepare  the shared/master aspect of  the job submission.
        Called  once  per  job  (both  split and  not-split).  If  the
        preparation contains some expensive actions it may be factored
        out in this method.

        Return value:
         - jobmasterconfig object understood by backend

        Arguments:
          - app : original application object
          - appmaster config : result of app.master_configure()

        """
        return None

    def prepare(self, app, appsubconfig, appmasterconfig, jobmasterconfig):
        """ Prepare the specific/subjob  aspect of the job submission.
        Called  once per  subjob if  splitting enabled.   If splitting
        disabled called  exactly once (the master  and specific aspect
        configured on the same job).

        Return value:
         - subjobconfig list of objects understood by backends

        Arguments:
          - app : original application object
          - appmaster config : result of app.master_configure()

         - appsubconfig : a list of results of app.configure() for each subjob
                         (or a master job if no splitting)
         - jobmasterconfig : a result of self.master_prepare()
        """

        raise NotImplementedError



# Bootstrap all of ganga, setup GPI, registries, etc.
from Ganga.Utility.Runtime import allRuntimes
from Ganga.Utility.Config import getConfig
from Ganga.Utility.logging import getLogger
from Ganga import _gangaPythonPath
import Ganga.Core
from Ganga.Core.GangaRepository import getRegistry

logger = getLogger(modulename=True)


def ganga_license():
    'Print the full license (GPL)'
    with open(os.path.join(_gangaPythonPath, '..', 'LICENSE_GPL')) as printable:
        logger.info(printable.read())


# ------------------------------------------------------------------------------------
# Setup the shutdown manager
from Ganga.Core.InternalServices import ShutdownManager
ShutdownManager.install()

## TODO need to implement loading of the config system properly here.
## loadPlugins and autoPopulateGPI will take this into account when loading objects

import ganga
from Ganga.Runtime import plugins

from Ganga.GPIDev.Base.Proxy import setProxyInterface
setProxyInterface(ganga)

from Ganga.Utility.Runtime import loadPlugins, autoPopulateGPI
loadPlugins(ganga)
autoPopulateGPI(ganga)

from Ganga.Core.GangaThread.WorkerThreads import startUpQueues
startUpQueues(ganga)

# ------------------------------------------------------------------------------------
# set the default value for the plugins

from Ganga.Utility.Runtime import setPluginDefaults
setPluginDefaults(ganga)


from Ganga.Runtime.bootstrap import manualExportToGPI
manualExportToGPI(ganga)


## Registries now add themselves to the Interface in this step
from Ganga.Runtime.Repository_runtime import startUpRegistries
startUpRegistries(ganga)

# ------------------------------------------------------------------------------------
#  bootstrap core modules
interactive = False
from Ganga.Core.GangaRepository import getRegistrySlice
Ganga.Core.bootstrap(getRegistrySlice('jobs'), interactive, my_interface=ganga)
Ganga.GPIDev.Lib.Config.bootstrap()

# ------------------------------------------------------------------------------------
# run post bootstrap hooks
for r in allRuntimes.values():
    try:
        r.postBootstrapHook()
    except Exception as err:
        logger.error("problems with post bootstrap hook for %s" % r.name)
        logger.error("Reason: %s" % str(err))




#!/usr/bin/env python
class AlgorithmError(Exception):

    """
    Class for Algorithm errors.
    """

    def __init__(self, message):
        self.message = message


class Algorithm(object):

    """
    Class to define user algorithm.
    """

    _attributes = ('results')

    def __init__(self):
        self.results = {}

    def __appendResult__(self, item, result):
        self.results[item] = result

    def process(self, item):
        """
        implements how to deal with the given data item. The output of the process can be
        appended to the self.results dictionary by calling self.__appendResult__(item, result). 

        @since: 0.0.1
        @author: Hurng-Chun Lee 
        @contact: hurngchunlee@gmail.com

        @param item is the sigle item from the data collection 
        @return True if the item is properly processed; otherwise False
        """
        raise NotImplementedError(
            'algorithm needs to be implemented to deal with the data item')

    def getResults(self):
        """
        returns the up-to-date results from what has been processed by this algorithm object.
        """
        return self.results
